import asyncio

from synchronicity import Synchronizer


async def run():
    try:
        while True:
            print("running")
            await asyncio.sleep(0.3)
    except asyncio.CancelledError:
        print("cancelled")
        await asyncio.sleep(0.1)
        print("handled cancellation")
        raise
    finally:
        await asyncio.sleep(0.1)
        print("exit async")


s = Synchronizer()
blocking_run = s.create_blocking(run)
try:
    blocking_run()
except KeyboardInterrupt:
    print("keyboard interrupt")



import asyncio
import time

from synchronicity import Synchronizer

s = Synchronizer()


@s.create_blocking
class MyClass:
    async def f(self, x):
        await asyncio.sleep(0.2)
        return x**2

    @s.nowrap
    def g(self, x):
        # This runs on the wrapped class
        return self.f(x) * x  # calls the blocking function


def test_nowrap():
    my_obj = MyClass()

    t0 = time.time()
    assert my_obj.f(111) == 12321
    assert 0.15 < time.time() - t0 < 0.25

    t0 = time.time()
    assert my_obj.g(111) == 1367631
    assert 0.15 < time.time() - t0 < 0.25



import asyncio
import asyncio.futures
import atexit
import collections.abc
import concurrent
import concurrent.futures
import contextlib
import functools
import inspect
import os
import platform
import threading
import types
import typing
import warnings
from typing import ForwardRef, Optional

import typing_extensions

from synchronicity.annotations import evaluated_annotation
from synchronicity.combined_types import FunctionWithAio, MethodWithAio

from .async_wrap import is_async_gen_function_follow_wrapped, is_coroutine_function_follow_wrapped, wraps_by_interface
from .callback import Callback
from .exceptions import UserCodeException, unwrap_coro_exception, wrap_coro_exception
from .interface import DEFAULT_CLASS_PREFIX, DEFAULT_FUNCTION_PREFIXES, Interface

_BUILTIN_ASYNC_METHODS = {
    "__aiter__": "__iter__",
    "__aenter__": "__enter__",
    "__aexit__": "__exit__",
    "__anext__": "__next__",
    "aclose": "close",
}

IGNORED_ATTRIBUTES = (
    # the "zope" lib monkey patches in some non-introspectable stuff on stdlib abc.ABC.
    # Ignoring __provides__ fixes an incompatibility with `channels[daphne]`,
    # where Synchronizer creation fails when wrapping contextlib._AsyncGeneratorContextManager
    "__provides__",
)

_RETURN_FUTURE_KWARG = "_future"

TARGET_INTERFACE_ATTR = "_sync_target_interface"
SYNCHRONIZER_ATTR = "_sync_synchronizer"


ASYNC_GENERIC_ORIGINS = (
    collections.abc.Awaitable,
    collections.abc.Coroutine,
    collections.abc.AsyncIterator,
    collections.abc.AsyncIterable,
    collections.abc.AsyncGenerator,
    contextlib.AbstractAsyncContextManager,
)


def _type_requires_aio_usage(annotation, declaration_module):
    if isinstance(annotation, ForwardRef):
        annotation = annotation.__forward_arg__
    if isinstance(annotation, str):
        try:
            annotation = evaluated_annotation(annotation, declaration_module=declaration_module)
        except Exception:
            # TODO: this will be incorrect in special case of `arg: "Awaitable[some_forward_ref_type]"`,
            #       but its a hard problem to solve without passing around globals everywhere
            return False

    if hasattr(annotation, "__origin__"):
        if annotation.__origin__ in ASYNC_GENERIC_ORIGINS:  # type: ignore
            return True
        # recurse for generic subtypes
        for a in getattr(annotation, "__args__", ()):
            if _type_requires_aio_usage(a, declaration_module):
                return True
    return False


def should_have_aio_interface(func):
    # determines if a blocking function gets an .aio attribute with an async interface to the function or not
    if is_coroutine_function_follow_wrapped(func) or is_async_gen_function_follow_wrapped(func):
        return True
    # check annotations if they contain any async entities that would need an event loop to be translated:
    # This catches things like vanilla functions returning Coroutines
    annos = getattr(func, "__annotations__", {})
    for anno in annos.values():
        if _type_requires_aio_usage(anno, func.__module__):
            return True
    return False


class Synchronizer:
    """Helps you offer a blocking (synchronous) interface to asynchronous code."""

    def __init__(
        self,
        multiwrap_warning=False,
        async_leakage_warning=True,
    ):
        self._multiwrap_warning = multiwrap_warning
        self._async_leakage_warning = async_leakage_warning
        self._loop = None
        self._loop_creation_lock = threading.Lock()
        self._thread = None
        self._owner_pid = None
        self._stopping = None

        if platform.system() == "Windows":
            # default event loop policy on windows spits out errors when
            # closing the event loop, so use WindowsSelectorEventLoopPolicy instead
            # https://stackoverflow.com/questions/45600579/asyncio-event-loop-is-closed-when-getting-loop
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        # Special attribute we use to go from wrapped <-> original
        self._wrapped_attr = "_sync_wrapped_%d" % id(self)
        self._original_attr = "_sync_original_%d" % id(self)

        # Special attribute to mark something as non-wrappable
        self._nowrap_attr = "_sync_nonwrap_%d" % id(self)
        self._input_translation_attr = "_sync_input_translation_%d" % id(self)
        self._output_translation_attr = "_sync_output_translation_%d" % id(self)

        # Prep a synchronized context manager in case one is returned and needs translation
        self._ctx_mgr_cls = contextlib._AsyncGeneratorContextManager
        self.create_blocking(self._ctx_mgr_cls)

        atexit.register(self._close_loop)

    _PICKLE_ATTRS = [
        "_multiwrap_warning",
        "_async_leakage_warning",
    ]

    def __getstate__(self):
        return dict([(attr, getattr(self, attr)) for attr in self._PICKLE_ATTRS])

    def __setstate__(self, d):
        for attr in self._PICKLE_ATTRS:
            setattr(self, attr, d[attr])

    def _start_loop(self):
        with self._loop_creation_lock:
            if self._loop and self._loop.is_running():
                # in case of a race between two _start_loop, the loop might already
                # be created here by another thread
                return self._loop

            is_ready = threading.Event()

            def thread_inner():
                async def loop_inner():
                    self._loop = asyncio.get_running_loop()
                    self._stopping = asyncio.Event()
                    is_ready.set()
                    await self._stopping.wait()  # wait until told to stop

                try:
                    asyncio.run(loop_inner())
                except RuntimeError as exc:
                    # Python 3.12 raises a RuntimeError when new threads are created at shutdown.
                    # Swallowing it here is innocuous, but ideally we will revisit this after
                    # refactoring the shutdown handlers that modal uses to avoid triggering it.
                    if "can't create new thread at interpreter shutdown" not in str(exc):
                        raise exc

            self._owner_pid = os.getpid()
            thread = threading.Thread(target=thread_inner, daemon=True)
            thread.start()
            is_ready.wait()  # TODO: this might block for a very short time
            self._thread = thread
            return self._loop

    def _close_loop(self):
        if self._thread is not None:
            if not self._loop.is_closed():
                # This also serves the purpose of waking up an idle loop
                self._loop.call_soon_threadsafe(self._stopping.set)
            self._thread.join()
            self._thread = None
            self._loop = None
            self._owner_pid = None

    def __del__(self):
        self._close_loop()

    def _get_loop(self, start=False) -> asyncio.AbstractEventLoop:
        if self._thread and not self._thread.is_alive():
            if self._owner_pid == os.getpid():
                # warn - thread died without us forking
                raise RuntimeError("Synchronizer thread unexpectedly died")

            self._thread = None
            self._loop = None

        if self._loop is None and start:
            return self._start_loop()
        return self._loop

    def _get_running_loop(self):
        # TODO: delete this method
        try:
            return asyncio.get_running_loop()
        except RuntimeError:
            return

    def _is_inside_loop(self):
        loop = self._get_loop()
        if loop is None:
            return False
        if threading.current_thread() != self._thread:
            # gevent does something bad that causes asyncio.get_running_loop() to return self._loop
            return False
        current_loop = self._get_running_loop()
        return loop == current_loop

    def _wrap_check_async_leakage(self, coro):
        """Check if a coroutine returns another coroutine (or an async generator) and warn.

        The reason this is important to catch is that otherwise even synchronized code might end up
        "leaking" async code into the caller.
        """
        if not self._async_leakage_warning:
            return coro

        async def coro_wrapped():
            value = await coro
            # TODO: we should include the name of the original function here
            if inspect.iscoroutine(value):
                warnings.warn(f"Potential async leakage: coroutine returned a coroutine {value}.")
            elif inspect.isasyncgen(value):
                warnings.warn(f"Potential async leakage: Coroutine returned an async generator {value}.")
            return value

        return coro_wrapped()

    def _wrap_instance(self, obj):
        # Takes an object and creates a new proxy object for it
        cls = obj.__class__
        cls_dct = cls.__dict__
        wrapper_cls = cls_dct[self._wrapped_attr][Interface.BLOCKING]
        new_obj = wrapper_cls.__new__(wrapper_cls)
        # Store a reference to the original object
        new_obj.__dict__[self._original_attr] = obj
        new_obj.__dict__[SYNCHRONIZER_ATTR] = self
        return new_obj

    def _translate_scalar_in(self, obj):
        # If it's an external object, translate it to the internal type
        if hasattr(obj, "__dict__"):
            if inspect.isclass(obj):  # TODO: functions?
                return obj.__dict__.get(self._original_attr, obj)
            else:
                return obj.__dict__.get(self._original_attr, obj)
        else:
            return obj

    def _translate_scalar_out(self, obj):
        # If it's an internal object, translate it to the external interface
        if inspect.isclass(obj):  # TODO: functions?
            cls_dct = obj.__dict__
            if self._wrapped_attr in cls_dct:
                return cls_dct[self._wrapped_attr][Interface.BLOCKING]
            else:
                return obj
        elif isinstance(obj, (typing.TypeVar, typing_extensions.ParamSpec)):
            if hasattr(obj, self._wrapped_attr):
                return getattr(obj, self._wrapped_attr)[Interface.BLOCKING]
            else:
                return obj
        else:
            cls_dct = obj.__class__.__dict__
            if self._wrapped_attr in cls_dct:
                # This is an *instance* of a synchronized class, translate its type
                return self._wrap(obj, interface=Interface.BLOCKING)
            else:
                return obj

    def _recurse_map(self, mapper, obj):
        if type(obj) == list:  # noqa: E721
            return list(self._recurse_map(mapper, item) for item in obj)
        elif type(obj) == tuple:  # noqa: E721
            return tuple(self._recurse_map(mapper, item) for item in obj)
        elif type(obj) == dict:  # noqa: E721
            return dict((key, self._recurse_map(mapper, item)) for key, item in obj.items())
        else:
            return mapper(obj)

    def _translate_in(self, obj):
        return self._recurse_map(self._translate_scalar_in, obj)

    def _translate_out(self, obj, interface=None):
        # TODO: remove deprecated interface arg - not used but needs deprecation path in case of external usage
        return self._recurse_map(lambda scalar: self._translate_scalar_out(scalar), obj)

    def _translate_coro_out(self, coro, original_func):
        async def unwrap_coro():
            res = await coro
            if getattr(original_func, self._output_translation_attr, True):
                return self._translate_out(res)
            return res

        return unwrap_coro()

    def _run_function_sync(self, coro, original_func):
        if self._is_inside_loop():
            raise Exception("Deadlock detected: calling a sync function from the synchronizer loop")

        coro = wrap_coro_exception(coro)
        coro = self._wrap_check_async_leakage(coro)
        loop = self._get_loop(start=True)

        coro_task = asyncio.ensure_future(coro, loop=loop)

        async def wrapper_coro():
            # this wrapper is needed since run_coroutine_threadsafe *only* accepts coroutines
            return await coro_task

        fut = asyncio.run_coroutine_threadsafe(wrapper_coro(), loop)
        try:
            while 1:
                try:
                    # poll every second to give Windows a chance to abort on Ctrl-C
                    #
                    value = fut.result(timeout=0.1)
                    break
                except concurrent.futures.TimeoutError:
                    pass
        except KeyboardInterrupt as exc:
            # in case there is a keyboard interrupt while we are waiting
            # we cancel the *underlying* coro_task (unlike what fut.cancel() would do)
            # and then wait for the *wrapper* coroutine to get a result back, which
            # happens after the cancellation resolves
            loop.call_soon_threadsafe(coro_task.cancel)
            try:
                value = fut.result()
            except concurrent.futures.CancelledError as expected_cancellation:
                # we *expect* this cancellation, but defer to the passed coro to potentially
                # intercept and treat the cancellation some other way
                expected_cancellation.__suppress_context__ = True
                raise exc  # if cancel - re-raise the original KeyboardInterrupt again

        if getattr(original_func, self._output_translation_attr, True):
            return self._translate_out(value)
        return value

    def _run_function_sync_future(self, coro, original_func):
        coro = wrap_coro_exception(coro)
        coro = self._wrap_check_async_leakage(coro)
        loop = self._get_loop(start=True)
        # For futures, we unwrap the result at this point, not in f_wrapped
        coro = unwrap_coro_exception(coro)
        coro = self._translate_coro_out(coro, original_func=original_func)
        return asyncio.run_coroutine_threadsafe(coro, loop)

    async def _run_function_async(self, coro, original_func):
        coro = wrap_coro_exception(coro)
        coro = self._wrap_check_async_leakage(coro)
        loop = self._get_loop(start=True)
        if self._is_inside_loop():
            value = await coro
        else:
            coro_task = asyncio.ensure_future(coro, loop=loop)

            async def run_coro():
                # this lets us cancel coro_task without cancelling
                # the c_fut directly, which would
                # to a_fut, regardless if `coro` handles the cancellation
                return await coro_task

            c_fut = asyncio.run_coroutine_threadsafe(run_coro(), loop)
            a_fut = asyncio.wrap_future(c_fut)

            try:
                # The shield here prevents a cancelled caller from cancelling c_fut directly
                # so that we can instead cancel the underlying coro_task and wait for it
                # to bubble up.
                # the loop + wait_for timeout is for windows ctrl-C compatibility since
                # windows doesn't truly interrupt the event loop on sigint
                while 1:
                    try:
                        value = await asyncio.wait_for(asyncio.shield(a_fut), timeout=0.1)
                        break
                    except asyncio.TimeoutError:
                        continue

            except asyncio.CancelledError:
                if a_fut.cancelled():
                    raise  # cancellation came from within c_fut
                loop.call_soon_threadsafe(coro_task.cancel)  # cancel inner task
                value = await a_fut  # typically also yields a cancellation error

        if getattr(original_func, self._output_translation_attr, True):
            return self._translate_out(value)
        return value

    def _run_generator_sync(self, gen, original_func):
        value, is_exc = None, False
        while True:
            try:
                if is_exc:
                    value = self._run_function_sync(gen.athrow(value), original_func)
                else:
                    value = self._run_function_sync(gen.asend(value), original_func)
            except UserCodeException as uc_exc:
                uc_exc.exc.__suppress_context__ = True
                raise uc_exc.exc
            except StopAsyncIteration:
                break
            try:
                value = yield value
                is_exc = False
            except BaseException as exc:
                value = exc
                is_exc = True

    async def _run_generator_async(self, gen, original_func):
        value, is_exc = None, False
        while True:
            try:
                if is_exc:
                    value = await self._run_function_async(gen.athrow(value), original_func)
                else:
                    value = await self._run_function_async(gen.asend(value), original_func)
            except UserCodeException as uc_exc:
                uc_exc.exc.__suppress_context__ = True
                raise uc_exc.exc
            except StopAsyncIteration:
                break
            try:
                value = yield value
                is_exc = False
            except BaseException as exc:
                value = exc
                is_exc = True

    def create_callback(self, f):
        return Callback(self, f)

    def _update_wrapper(self, f_wrapped, f, name=None, interface=None, target_module=None):
        """Very similar to functools.update_wrapper"""
        functools.update_wrapper(f_wrapped, f)
        if name is not None:
            f_wrapped.__name__ = name
            f_wrapped.__qualname__ = name
        if target_module is not None:
            f_wrapped.__module__ = target_module
        setattr(f_wrapped, SYNCHRONIZER_ATTR, self)
        setattr(f_wrapped, TARGET_INTERFACE_ATTR, interface)

    def _wrap_callable(
        self,
        f,
        interface,
        name=None,
        allow_futures=True,
        unwrap_user_excs=True,
        target_module=None,
        include_aio_interface=True,
    ):
        if hasattr(f, self._original_attr):
            if self._multiwrap_warning:
                warnings.warn(f"Function {f} is already wrapped, but getting wrapped again")
            return f

        if name is None:
            _name = DEFAULT_FUNCTION_PREFIXES[interface] + f.__name__
        else:
            _name = name

        @wraps_by_interface(interface, f)
        def f_wrapped(*args, **kwargs):
            return_future = kwargs.pop(_RETURN_FUTURE_KWARG, False)

            # If this gets called with an argument that represents an external type,
            # translate it into an internal type
            if getattr(f, self._input_translation_attr, True):
                args = self._translate_in(args)
                kwargs = self._translate_in(kwargs)

            # Call the function
            res = f(*args, **kwargs)

            # Figure out if this is a coroutine or something
            is_coroutine = inspect.iscoroutine(res)
            is_asyncgen = inspect.isasyncgen(res)

            if return_future:
                if not allow_futures:
                    raise Exception("Can not return future for this function")
                elif is_coroutine:
                    return self._run_function_sync_future(res, f)
                elif is_asyncgen:
                    raise Exception("Can not return futures for generators")
                else:
                    return res
            elif is_coroutine:
                if interface == Interface._ASYNC_WITH_BLOCKING_TYPES:
                    coro = self._run_function_async(res, f)
                    if not is_coroutine_function_follow_wrapped(f):
                        # If this is a non-async function that returns a coroutine,
                        # then this is the exit point, and we need to unwrap any
                        # wrapped exception here. Otherwise, the exit point is
                        # in async_wrap.py
                        coro = unwrap_coro_exception(coro)
                    return coro
                elif interface == Interface.BLOCKING:
                    # This is the exit point, so we need to unwrap the exception here
                    try:
                        return self._run_function_sync(res, f)
                    except StopAsyncIteration:
                        # this is a special case for handling __next__ wrappers around
                        # __anext__ that raises StopAsyncIteration
                        raise StopIteration()
                    except UserCodeException as uc_exc:
                        # Used to skip a frame when called from `proxy_method`.
                        if unwrap_user_excs and not (Interface.BLOCKING and include_aio_interface):
                            uc_exc.exc.__suppress_context__ = True
                            raise uc_exc.exc
                        else:
                            raise uc_exc
            elif is_asyncgen:
                # Note that the _run_generator_* functions handle their own
                # unwrapping of exceptions (this happens during yielding)
                if interface == Interface._ASYNC_WITH_BLOCKING_TYPES:
                    return self._run_generator_async(res, f)
                elif interface == Interface.BLOCKING:
                    return self._run_generator_sync(res, f)
            else:
                if inspect.isfunction(res) or isinstance(res, functools.partial):  # TODO: HACKY HACK
                    # TODO: this is needed for decorator wrappers that returns functions
                    # Maybe a bit of a hacky special case that deserves its own decorator
                    @wraps_by_interface(interface, res)
                    def f_wrapped(*args, **kwargs):
                        args = self._translate_in(args)
                        kwargs = self._translate_in(kwargs)
                        f_res = res(*args, **kwargs)
                        if getattr(f, self._output_translation_attr, True):
                            return self._translate_out(f_res)
                        else:
                            return f_res

                    return f_wrapped

                if getattr(f, self._output_translation_attr, True):
                    return self._translate_out(res, interface)
                else:
                    return res

        self._update_wrapper(f_wrapped, f, _name, interface, target_module=target_module)
        setattr(f_wrapped, self._original_attr, f)

        if interface == Interface.BLOCKING and include_aio_interface and should_have_aio_interface(f):
            # special async interface
            # this async interface returns *blocking* instances of wrapped objects, not async ones:
            async_interface = self._wrap_callable(
                f,
                interface=Interface._ASYNC_WITH_BLOCKING_TYPES,
                name=name,
                allow_futures=allow_futures,
                unwrap_user_excs=unwrap_user_excs,
                target_module=target_module,
            )
            f_wrapped = FunctionWithAio(f_wrapped, async_interface, self)
            self._update_wrapper(f_wrapped, f, _name, interface, target_module=target_module)
            setattr(f_wrapped, self._original_attr, f)

        return f_wrapped

    def _wrap_proxy_method(
        synchronizer_self,
        method,
        interface,
        allow_futures=True,
        include_aio_interface=True,
    ):
        if getattr(method, synchronizer_self._nowrap_attr, None):
            # This method is marked as non-wrappable
            return method

        wrapped_method = synchronizer_self._wrap_callable(
            method,
            interface,
            allow_futures=allow_futures,
            unwrap_user_excs=False,
        )

        @wraps_by_interface(interface, wrapped_method)
        def proxy_method(self, *args, **kwargs):
            instance = self.__dict__[synchronizer_self._original_attr]
            try:
                return wrapped_method(instance, *args, **kwargs)
            except UserCodeException as uc_exc:
                uc_exc.exc.__suppress_context__ = True
                raise uc_exc.exc

        if interface == Interface.BLOCKING and include_aio_interface and should_have_aio_interface(method):
            async_proxy_method = synchronizer_self._wrap_proxy_method(
                method, Interface._ASYNC_WITH_BLOCKING_TYPES, allow_futures
            )
            return MethodWithAio(proxy_method, async_proxy_method, synchronizer_self)

        return proxy_method

    def _wrap_proxy_staticmethod(self, method, interface):
        orig_function = method.__func__
        method = self._wrap_callable(orig_function, interface)
        if isinstance(method, FunctionWithAio):
            return method  # no need to wrap a FunctionWithAio in a staticmethod, as it won't get bound anyways
        return staticmethod(method)

    def _wrap_proxy_classmethod(self, orig_classmethod, interface):
        orig_func = orig_classmethod.__func__
        method = self._wrap_callable(orig_func, interface, include_aio_interface=False)

        if interface == Interface.BLOCKING and should_have_aio_interface(orig_func):
            async_method = self._wrap_callable(orig_func, Interface._ASYNC_WITH_BLOCKING_TYPES)
            return MethodWithAio(method, async_method, self, is_classmethod=True)

        return classmethod(method)

    def _wrap_proxy_property(self, prop, interface):
        kwargs = {}
        for attr in ["fget", "fset", "fdel"]:
            if getattr(prop, attr):
                func = getattr(prop, attr)
                kwargs[attr] = self._wrap_proxy_method(
                    func, interface, allow_futures=False, include_aio_interface=False
                )
        return property(**kwargs)

    def _wrap_proxy_constructor(synchronizer_self, cls, interface):
        """Returns a custom __init__ for the subclass."""

        def my_init(self, *args, **kwargs):
            # Create base instance
            args = synchronizer_self._translate_in(args)
            kwargs = synchronizer_self._translate_in(kwargs)
            instance = cls(*args, **kwargs)

            # Register self as the wrapped one
            interface_instances = {interface: self}
            instance.__dict__[synchronizer_self._wrapped_attr] = interface_instances

            # Store a reference to the original object
            self.__dict__[synchronizer_self._original_attr] = instance

        synchronizer_self._update_wrapper(my_init, cls.__init__, interface=interface)
        setattr(my_init, synchronizer_self._original_attr, cls.__init__)
        return my_init

    def _wrap_class(self, cls, interface, name, target_module=None):
        new_bases = []
        for base in cls.__dict__.get("__orig_bases__", cls.__bases__):
            base_is_generic = hasattr(base, "__origin__")
            if base is object or (base_is_generic and base.__origin__ == typing.Generic):
                new_bases.append(base)  # no need to wrap these, just add them as base classes
            else:
                if base_is_generic:
                    wrapped_generic = self._wrap(base.__origin__, interface, require_already_wrapped=(name is not None))
                    new_bases.append(wrapped_generic.__class_getitem__(base.__args__))
                else:
                    new_bases.append(self._wrap(base, interface, require_already_wrapped=(name is not None)))

        bases = tuple(new_bases)
        new_dict = {self._original_attr: cls}
        if cls is not None:
            new_dict["__init__"] = self._wrap_proxy_constructor(cls, interface)

        for k, v in cls.__dict__.items():
            if k in _BUILTIN_ASYNC_METHODS:
                k_sync = _BUILTIN_ASYNC_METHODS[k]
                new_dict[k_sync] = self._wrap_proxy_method(
                    v,
                    Interface.BLOCKING,
                    allow_futures=False,
                    include_aio_interface=False,
                )
                new_dict[k] = self._wrap_proxy_method(
                    v,
                    Interface._ASYNC_WITH_BLOCKING_TYPES,
                    allow_futures=False,
                )
            elif k in ("__new__", "__init__"):
                # Skip custom constructor in the wrapped class
                # Instead, delegate to the base class constructor and wrap it
                pass
            elif k in IGNORED_ATTRIBUTES:
                pass
            elif isinstance(v, staticmethod):
                # TODO(erikbern): this feels pretty hacky
                new_dict[k] = self._wrap_proxy_staticmethod(v, Interface.BLOCKING)
            elif isinstance(v, classmethod):
                new_dict[k] = self._wrap_proxy_classmethod(v, Interface.BLOCKING)
            elif isinstance(v, property):
                new_dict[k] = self._wrap_proxy_property(v, Interface.BLOCKING)
            elif isinstance(v, MethodWithAio):
                # if library defines its own MethodWithAio descriptor we transfer it "as is" to the wrapper
                # without wrapping it again
                new_dict[k] = v
            elif callable(v):
                new_dict[k] = self._wrap_proxy_method(v, Interface.BLOCKING)

        if name is None:
            name = DEFAULT_CLASS_PREFIX + cls.__name__

        new_cls = types.new_class(name, bases, exec_body=lambda ns: ns.update(new_dict))
        new_cls.__module__ = cls.__module__ if target_module is None else target_module
        new_cls.__doc__ = cls.__doc__
        if "__annotations__" in cls.__dict__:
            new_cls.__annotations__ = cls.__annotations__  # transfer annotations

        setattr(new_cls, SYNCHRONIZER_ATTR, self)
        return new_cls

    def _wrap(
        self,
        obj,
        interface,
        name=None,
        require_already_wrapped=False,
        target_module=None,
    ):
        # This method works for classes, functions, and instances
        # It wraps the object, and caches the wrapped object

        # Get the list of existing interfaces
        if hasattr(obj, "__dict__"):
            if self._wrapped_attr not in obj.__dict__:
                if isinstance(obj.__dict__, dict):
                    # This works for instances
                    obj.__dict__.setdefault(self._wrapped_attr, {})
                else:
                    # This works for classes & functions
                    setattr(obj, self._wrapped_attr, {})
            interfaces = obj.__dict__[self._wrapped_attr]
        else:
            # e.g., TypeVar in Python>=3.12
            if not hasattr(obj, self._wrapped_attr):
                setattr(obj, self._wrapped_attr, {})
            interfaces = getattr(obj, self._wrapped_attr)

        # If this is already wrapped, return the existing interface
        if interface in interfaces:
            if self._multiwrap_warning:
                warnings.warn(f"Object {obj} is already wrapped, but getting wrapped again")
            return interfaces[interface]

        if require_already_wrapped:
            # This happens if a class has a custom name but its base class doesn't
            raise RuntimeError(f"{obj} needs to be serialized explicitly with a custom name")

        # Wrap object (different cases based on the type)
        if inspect.isclass(obj):
            new_obj = self._wrap_class(
                obj,
                interface,
                name,
                target_module=target_module,
            )
        elif inspect.isfunction(obj):
            new_obj = self._wrap_callable(obj, interface, name, target_module=target_module)
        elif isinstance(obj, typing_extensions.ParamSpec):
            new_obj = self._wrap_param_spec(obj, interface, name, target_module)
        elif isinstance(obj, typing.TypeVar):
            new_obj = self._wrap_type_var(obj, interface, name, target_module)
        elif self._wrapped_attr in obj.__class__.__dict__:
            new_obj = self._wrap_instance(obj)
        else:
            raise Exception("Argument %s is not a class or a callable" % obj)

        # Store the interface on the obj and return
        interfaces[interface] = new_obj
        return new_obj

    def _wrap_type_var(self, obj, interface, name, target_module):
        # TypeVar translation is needed only for type stub generation, in case the
        # "bound" attribute refers to a translatable type.

        # Creates a new identical TypeVar, marked with synchronicity's special attributes
        # This lets type stubs "translate" the `bounds` attribute on emitted type vars
        # if picked up from module scope and in generics using the base implementation type

        # TODO(elias): Refactor - since this isn't used for live apps, move type stub generation into genstub
        new_obj = typing.TypeVar(name, bound=obj.__bound__)  # noqa
        setattr(new_obj, self._original_attr, obj)
        setattr(new_obj, SYNCHRONIZER_ATTR, self)
        setattr(new_obj, TARGET_INTERFACE_ATTR, interface)
        new_obj.__module__ = target_module
        if not hasattr(obj, self._wrapped_attr):
            setattr(obj, self._wrapped_attr, {})
        getattr(obj, self._wrapped_attr)[interface] = new_obj
        return new_obj

    def _wrap_param_spec(self, obj, interface, name, target_module):
        # TODO(elias): Refactor - since this isn't used for live apps, move type stub generation into genstub
        new_obj = typing_extensions.ParamSpec(name)  # noqa
        setattr(new_obj, self._original_attr, obj)
        setattr(new_obj, SYNCHRONIZER_ATTR, self)
        setattr(new_obj, TARGET_INTERFACE_ATTR, interface)
        new_obj.__module__ = target_module
        if not hasattr(obj, self._wrapped_attr):
            setattr(obj, self._wrapped_attr, {})
        getattr(obj, self._wrapped_attr)[interface] = new_obj
        return new_obj

    def nowrap(self, obj):
        setattr(obj, self._nowrap_attr, True)
        return obj

    def no_input_translation(self, obj):
        setattr(obj, self._input_translation_attr, False)
        return obj

    def no_output_translation(self, obj):
        setattr(obj, self._output_translation_attr, False)
        return obj

    def no_io_translation(self, obj):
        return self.no_input_translation(self.no_output_translation(obj))

    # New interface that (almost) doesn't mutate objects
    def create_blocking(self, obj, name: Optional[str] = None, target_module: Optional[str] = None):
        wrapped = self._wrap(obj, Interface.BLOCKING, name, target_module=target_module)
        return wrapped

    def is_synchronized(self, obj):
        if inspect.isclass(obj) or inspect.isfunction(obj):
            return hasattr(obj, self._original_attr)
        else:
            return hasattr(obj.__class__, self._original_attr)



"""
Improvement Ideas:
* Extract this into its own package, not linked to synchronicity, but with good extension plugs?
* Don't use the wrapped synchronicity types directly, and instead emit stubs based on the root
  implementation types directly (but translated to blocking).
* Let synchronicity emit actual function bodies, to avoid runtime wrapping altogether
"""

import collections
import collections.abc
import contextlib
import contextvars
import enum
import importlib
import inspect
import sys
import typing
from logging import getLogger
from pathlib import Path
from typing import TypeVar
from unittest import mock

import sigtools.specifiers  # type: ignore
import typing_extensions
from sigtools._signatures import EmptyAnnotation, UpgradedAnnotation, UpgradedParameter  # type: ignore

import synchronicity
from synchronicity import combined_types, overload_tracking
from synchronicity.annotations import TYPE_CHECKING_OVERRIDES, evaluated_annotation
from synchronicity.async_wrap import is_coroutine_function_follow_wrapped
from synchronicity.interface import Interface
from synchronicity.synchronizer import (
    SYNCHRONIZER_ATTR,
    TARGET_INTERFACE_ATTR,
    FunctionWithAio,
    MethodWithAio,
)

logger = getLogger(__name__)


def safe_get_module(obj: typing.Any) -> typing.Optional[str]:
    """Handles some special cases where obj.__module__ isn't correct or ugly

    e.g. in Python 3.8 contextvars.ContextVar.__module__ == "builtins"
    and in Python 3.11 contextvars.ContextVar.__module__ == "_contextvars"
    and in emitted code it should *preferably* be "contextvars"
    """
    if obj == contextvars.ContextVar:
        return "contextvars"

    if not hasattr(obj, "__module__"):
        return None

    if obj.__module__ in ("_contextvars", "_asyncio"):
        return obj.__module__[1:]  # strip leading underscore

    try:
        if (obj.__module__, obj.__name__) == ("typing", "Concatenate"):
            # typing_extensions.Concatenate forwards typing.Concatenate if
            # available, but we still want to emit typing_extensions to be
            # backwards compatible
            return "typing_extensions"
    except Exception:
        pass

    return obj.__module__


def generic_copy_with_args(specific_type, new_args):
    origin = typing_extensions.get_origin(specific_type)
    if hasattr(specific_type, "copy_with") and origin not in (typing.Callable, collections.abc.Callable):
        # not strictly necessary, but this makes the type stubs
        # preserve generic alias names when possible, e.g. using `typing.Iterator`
        # instead of changing it into `collections.abc.Iterator`
        return specific_type.copy_with(new_args)

    return origin[new_args]


def add_prefix_arg(arg_name, remove_args=0):
    def inject_arg_func(sig: inspect.Signature):
        parameters = list(sig.parameters.values())
        return sig.replace(
            parameters=[
                UpgradedParameter(arg_name, inspect.Parameter.POSITIONAL_OR_KEYWORD),
                *parameters[remove_args:],
            ]
        )

    return inject_arg_func


def replace_type_vars(replacement_dict: typing.Dict[type, type]):
    def _replace_type_vars_rec(tp: typing.Type[typing.Any]):
        origin = getattr(tp, "__origin__", None)
        args = safe_get_args(tp)

        if isinstance(tp, (typing_extensions.ParamSpecArgs, typing_extensions.ParamSpecKwargs)):
            new_origin_type_var = _replace_type_vars_rec(origin)
            return type(tp)(new_origin_type_var)

        if type(tp) is list:  # typically first argument to typing.Callable
            # intentionally not using isinstance, since ParamSpec is a subclass of list in Python 3.9
            return [_replace_type_vars_rec(arg) for arg in tp]

        if tp in replacement_dict:
            return replacement_dict[tp]

        if origin:
            newargs = tuple(_replace_type_vars_rec(a) for a in args)
            return generic_copy_with_args(tp, newargs)

        return tp

    def _replace_type_vars_in_sig(sig: inspect.Signature):
        parameters = [p.replace(annotation=_replace_type_vars_rec(p.annotation)) for p in sig.parameters.values()]
        return sig.replace(
            parameters=parameters,
            return_annotation=_replace_type_vars_rec(sig.return_annotation),
        )

    return _replace_type_vars_in_sig


def _get_type_vars(typ, synchronizer, home_module):
    origin = getattr(typ, "__origin__", None)  # typing.get_origin returns None for ParamSpecArgs on <=Python3.9
    ret = set()
    if isinstance(typ, typing.TypeVar):
        # check if it's translated (due to bounds= attributes etc.)
        typ = synchronizer._translate_out(typ)
        ret.add(typ)
    elif isinstance(typ, (typing_extensions.ParamSpecArgs, typing_extensions.ParamSpecKwargs)):
        param_spec = origin
        param_spec = synchronizer._translate_out(param_spec)
        ret.add(param_spec)
    elif origin:
        for arg in safe_get_args(typ):
            ret |= _get_type_vars(arg, synchronizer, home_module)
    else:
        # Copied string annotation handling from StubEmitter.translate_annotations - TODO: unify?
        # The reason it cant be used directly is that this method should return the original type params
        # and not translated values
        if isinstance(typ, typing.ForwardRef):  # TypeVars wrap their arguments as ForwardRefs (sometimes?)
            if hasattr(typ, "__forward_module__") and typ.__forward_module__ is not None:
                return ret
            typ = typ.__forward_arg__
        if isinstance(typ, str):
            try:
                typ = evaluated_annotation(typ, declaration_module=home_module)
            except Exception:
                logger.exception(f"Error when evaluating {typ} in {home_module}. Falling back to string typ")
                return ret
            return _get_type_vars(typ, synchronizer, home_module)
    return ret


def _get_func_type_vars(func, synchronizer: synchronicity.Synchronizer) -> typing.Set[type]:
    ret = set()
    home_module = safe_get_module(func)
    for typ in getattr(func, "__annotations__", {}).values():
        ret |= _get_type_vars(typ, synchronizer, home_module)
    return ret


def safe_get_args(annotation):
    # "polyfill" of Python 3.10+ typing.get_args() behavior of
    # not putting ParamSpec and Ellipsis in a list when used as first argument to a Callable
    # This can be removed if we drop support for *generating type stubs using Python <=3.9*
    args = typing.get_args(annotation)
    if sys.version_info[:2] <= (3, 9) and typing.get_origin(annotation) == collections.abc.Callable:
        if (
            args
            and type(args[0]) is list
            and args[0]
            and isinstance(args[0][0], (typing_extensions.ParamSpec, type(...)))
        ):
            args = (args[0][0],) + args[1:]

    return args


def get_specific_generic_name(annotation):
    """get the name of the generic type of a "specific" type (with args)
    e.g.
    >>> get_specific_generic_name(typing.List[str])
    "List"
    """
    if hasattr(annotation, "__name__"):
        # this works on new pythons
        return annotation.__name__
    elif hasattr(annotation, "_name") and annotation._name is not None:
        # fallback for older Python (at least 3.8)
        return annotation._name
    else:
        # also an old python
        return get_specific_generic_name(annotation.__origin__)


class StubEmitter:
    def __init__(self, target_module):
        self.target_module = target_module
        self.imports = set()
        self.parts = []
        self._indentation = "    "
        self.global_types = set()
        self.referenced_global_types = set()
        self._typevar_inner_replacements = {}

    @classmethod
    def from_module(cls, module):
        emitter = cls(module.__name__)
        explicit_members = module.__dict__.get("__all__", [])
        for entity_name, entity in module.__dict__.copy().items():
            if (
                hasattr(entity, "__module__")
                and safe_get_module(entity) != module.__name__
                and entity_name not in explicit_members
                and typing.get_origin(entity) is not typing.Literal
            ):
                continue  # skip imported stuff, unless it's explicitly in __all__
            if inspect.isclass(entity):
                emitter.add_class(entity, entity_name)
            elif inspect.isfunction(entity) or isinstance(entity, FunctionWithAio):
                emitter.add_function(entity, entity_name)
            elif isinstance(entity, (typing.TypeVar, typing_extensions.ParamSpec)):
                emitter.add_type_var(entity, entity_name)
            elif hasattr(entity, "__class__") and safe_get_module(entity.__class__) == module.__name__:
                # instances of stuff
                emitter.add_variable(entity.__class__, entity_name)
            elif typing.get_origin(entity) is typing.Literal:
                emitter.add_literal(entity, entity_name)

        for varname, annotation in getattr(module, "__annotations__", {}).items():
            emitter.add_variable(annotation, varname)

        return emitter

    def add_variable(self, annotation, name):
        # TODO: evaluate string annotations
        self.parts.append(self._get_var_annotation(name, annotation))

    def add_literal(self, entity, name):
        self.parts.append(f"{name} = {str(entity)}")

    def add_function(self, func, name, indentation_level=0):
        # adds function source code to module
        if isinstance(func, FunctionWithAio):
            # this is a synchronicity-emitted replacement function/method for an originally async function
            self.parts.append(self._get_dual_function_source(func, name, indentation_level))
        else:
            self.parts.append(self._get_function_source_with_overloads(func, name, indentation_level))

    def _get_translated_class_bases(self, cls):
        bases = []
        for b in cls.__dict__.get("__orig_bases__", cls.__bases__):
            bases.append(self._translate_global_annotation(b, cls))
        return bases

    def add_class(self, cls, name) -> None:
        self.global_types.add(name)

        if issubclass(cls, enum.Enum):
            # Do not translate Enum classes.
            self.imports.add("enum")
            self.parts.append(inspect.getsource(cls))
            return

        bases = []
        generic_type_vars: typing.Set[type] = set()
        for b in self._get_translated_class_bases(cls):
            if b is not object:
                bases.append(self._formatannotation(b))
            if getattr(b, "__origin__", None) == typing.Generic:
                generic_type_vars |= {a for a in b.__args__}

        bases_str = "" if not bases else "(" + ", ".join(bases) + ")"
        decl = f"class {name}{bases_str}:"
        var_annotations = []
        methods = []

        annotations = cls.__dict__.get("__annotations__", {})
        annotations = {k: self._translate_global_annotation(annotation, cls) for k, annotation in annotations.items()}

        body_indent_level = 1
        body_indent = self._indent(body_indent_level)

        for varname, annotation in annotations.items():
            var_annotations.append(f"{body_indent}{self._get_var_annotation(varname, annotation)}")
        if var_annotations:
            var_annotations.append("")  # formatting ocd - add an extra newline after var annotations

        for entity_name, entity in cls.__dict__.items():
            if inspect.isfunction(entity):
                methods.append(self._get_function_source_with_overloads(entity, entity_name, body_indent_level))

            elif isinstance(entity, classmethod):
                fn_source = self._get_function_source_with_overloads(entity.__func__, entity_name, body_indent_level)
                methods.append(f"{body_indent}@classmethod\n{fn_source}")

            elif isinstance(entity, staticmethod):
                fn_source = self._get_function_source_with_overloads(entity.__func__, entity_name, body_indent_level)
                methods.append(f"{body_indent}@staticmethod\n{fn_source}")

            elif isinstance(entity, property):
                fn_source = self._get_function_source_with_overloads(entity.fget, entity_name, body_indent_level)
                methods.append(f"{body_indent}@property\n{fn_source}")

                if entity.fset:
                    fn_source = self._get_function_source_with_overloads(entity.fset, entity_name, body_indent_level)
                    methods.append(f"{body_indent}@{entity_name}.setter\n{fn_source}")

                if entity.fdel:
                    fn_source = self._get_function_source_with_overloads(entity.fdel, entity_name, body_indent_level)
                    methods.append(f"{body_indent}@{entity_name}.deleter\n{fn_source}")

            elif isinstance(entity, FunctionWithAio):
                # Note: FunctionWithAio is used for staticmethods
                methods.append(
                    self._get_dual_function_source(
                        entity, entity_name, body_indent_level, parent_generic_type_vars=generic_type_vars
                    )
                )
            elif isinstance(entity, MethodWithAio):
                if entity._is_classmethod:
                    # Classmethods with type vars on the cls variable don't work with "dual interface functions"
                    # at the moment, so we only output a stub for the blocking interface
                    # TODO(elias): allow dual type stubs as long as no type vars are being used in the class var
                    fn_source = self._get_function_source_with_overloads(entity._func, entity_name, body_indent_level)
                    src = f"{body_indent}@classmethod\n{fn_source}"
                else:
                    src = self._get_dual_function_source(
                        entity, entity_name, body_indent_level, parent_generic_type_vars=generic_type_vars
                    )
                methods.append(src)

        padding = [] if var_annotations or methods else [f"{body_indent}..."]
        self.parts.append(
            "\n".join(
                [
                    decl,
                    *var_annotations,
                    *methods,
                    *padding,
                ]
            )
        )

    def _get_dual_function_source(
        self,
        entity: typing.Union[MethodWithAio, FunctionWithAio],
        entity_name,
        body_indent_level,
        parent_generic_type_vars: typing.Set[type] = set(),  # if a method of a Generic class - the set of type vars
    ) -> str:
        if isinstance(entity, FunctionWithAio):
            transform_signature = add_prefix_arg(
                "self"
            )  # signature is moved into a protocol class, so we need a self where there previously was none
        elif entity._is_classmethod:
            # TODO: dual protocol for classmethods having annotated cls attributes
            raise Exception("Not supported")
        else:
            transform_signature = add_prefix_arg("self", 1)
        # Emits type stub for a "dual" function that is both callable and has an .aio callable with an async version
        # Currently this is emitted as a typing.Protocol declaration + instance with a __call__ and aio method
        self.imports.add("typing_extensions")
        # Synchronicity specific blocking + async method
        body_indent = self._indent(body_indent_level)

        (
            typevar_signature_transform,
            parent_type_var_names_spec,
            protocol_declaration_type_var_spec,
        ) = self._prepare_method_generic_type_vars(entity, parent_generic_type_vars)

        def final_transform_signature(sig):
            return typevar_signature_transform(transform_signature(sig))

        # create an inline protocol type, inlining both the blocking and async interfaces:
        blocking_func_source = self._get_function_source_with_overloads(
            entity._func,
            "__call__",
            body_indent_level + 1,
            transform_signature=final_transform_signature,
        )
        aio_func_source = self._get_function_source_with_overloads(
            entity._aio_func,
            "aio",
            body_indent_level + 1,
            transform_signature=final_transform_signature,
        )

        protocol_attr = f"""\
{body_indent}class __{entity_name}_spec(typing_extensions.Protocol{protocol_declaration_type_var_spec}):
{blocking_func_source}
{aio_func_source}
{body_indent}{entity_name}: __{entity_name}_spec{parent_type_var_names_spec}
"""

        return protocol_attr

    def _prepare_method_generic_type_vars(self, entity, parent_generic_type_vars):
        # Check any Generic TypeVar/ParamSpec used in the class x method, in order to
        # create a new type var for the protocol itself, since a "namespaced class" can't use the
        # generic type vars of its "parent class" directly. This will roughly translate to:
        # T = TypeVar("T")
        # T_INNER = TypeVar("T_INNER")
        # class Foo(Generic[T]):
        #     class Method(typing.Protocol[T_INNER]):
        #         def __call__(self, t: T_INNER):
        #             ...
        #
        #     method: Method[T]
        func_type_vars = _get_func_type_vars(entity._func, entity._synchronizer)
        typevar_overlap = parent_generic_type_vars & func_type_vars

        for tvar in typevar_overlap:
            if tvar in self._typevar_inner_replacements:
                continue
            replacement_typevar_name = tvar.__name__ + "_INNER"
            if isinstance(tvar, typing_extensions.ParamSpec):
                new_tvar = typing_extensions.ParamSpec(replacement_typevar_name)  # type: ignore
            else:
                new_tvar = typing.TypeVar(replacement_typevar_name, covariant=True)  # type: ignore
            new_tvar.__module__ = self.target_module  # avoid referencing synchronicity.type_stubs
            self._typevar_inner_replacements[tvar] = new_tvar
            self.add_type_var(new_tvar, replacement_typevar_name)  # type: ignore

        extra_instance_args = []
        extra_declaration_args = []

        if isinstance(entity, MethodWithAio):
            # support for typing.Self (which would otherwise reference the protocol class)
            superself_name = "SUPERSELF"
            superself_var = typing.TypeVar(superself_name, covariant=True)  # type: ignore
            superself_var.__module__ = self.target_module
            self.add_type_var(superself_var, superself_name)
            self._typevar_inner_replacements[typing_extensions.Self] = superself_var
            self.imports.add("typing_extensions")
            extra_instance_args = ["typing_extensions.Self"]
            extra_declaration_args = ["SUPERSELF"]

        protocol_generic_args = [
            self._typevar_inner_replacements[tvar].__name__ for tvar in typevar_overlap
        ] + extra_declaration_args
        if protocol_generic_args:
            original_type_var_names = []
            for tvar in typevar_overlap:
                original_type_var_names.append(self._formatannotation(tvar))

            instance_argstr = ", ".join(original_type_var_names + extra_instance_args)
            parent_type_var_names_spec = f"[{instance_argstr}]"
            declaration_argstr = ", ".join(protocol_generic_args)
            protocol_declaration_type_var_spec = f"[{declaration_argstr}]"

            # recursively replace any used type vars in the function annotation with newly created
            transform_signature = replace_type_vars(self._typevar_inner_replacements)
        else:
            transform_signature = lambda x: x  # noqa
            parent_type_var_names_spec = ""
            protocol_declaration_type_var_spec = ""

        return transform_signature, parent_type_var_names_spec, protocol_declaration_type_var_spec

    def add_type_var(self, type_var: typing.Union[typing.TypeVar, typing_extensions.ParamSpec], name: str):
        if name in self.global_types:
            # skip already added type
            # TODO: check that the already added type is the same?
            return

        if isinstance(type_var, typing_extensions.ParamSpec):
            type_module = "typing_extensions"  # this ensures stubs created by newer Python's still work on Python 3.9
            type_name = "ParamSpec"
        elif isinstance(type_var, typing.TypeVar):
            type_module = "typing"
            type_name = "TypeVar"
        else:
            raise TypeError("Not a TypeVar/ParamSpec")

        self.imports.add(type_module)
        args = [f'"{name}"']
        if type_var.__bound__ and type_var.__bound__ is not type(None):
            translated_bound = self._translate_global_annotation(type_var.__bound__, type_var)
            str_annotation = self._formatannotation(translated_bound)
            args.append(f'bound="{str_annotation}"')
        if isinstance(type_var, typing.TypeVar) and type_var.__covariant__:
            args.append("covariant=True")

        self.global_types.add(name)
        self.parts.append(f"{name} = {type_module}.{type_name}({', '.join(args)})")

    def get_source(self):
        missing_types = self.referenced_global_types - self.global_types
        if missing_types:
            print(f"WARNING: {self.target_module} missing the following referenced types, expected to be in module")
            for t in missing_types:
                print(t)
        import_src = "\n".join(sorted(f"import {mod}" for mod in self.imports))
        stubs = "\n\n".join(self.parts)
        return f"{import_src}\n\n{stubs}".lstrip()

    def _ensure_import(self, typ):
        # add import for a single type, non-recursive (See _register_imports)
        # also marks the type name as directly referenced if it's part of the target module
        # so we can sanity check
        module = safe_get_module(typ)

        if module not in (self.target_module, "builtins"):
            self.imports.add(module)

        if module == self.target_module:
            if not hasattr(typ, "__name__"):
                # weird special case with Generic subclasses in the target module
                # fall back to the origin name
                generic_origin = typ.__origin__
                name = generic_origin.__name__
            else:
                name = typ.__name__
            self.referenced_global_types.add(name)

    def _register_imports(self, type_annotation):
        # recursively makes sure a type and any of its type arguments (for generics) are imported
        origin = getattr(type_annotation, "__origin__", None)
        if origin is None:
            # "scalar" base type
            if hasattr(type_annotation, "__module__"):
                self._ensure_import(type_annotation)
            return

        self._ensure_import(type_annotation)  # import the generic itself's module
        for arg in getattr(type_annotation, "__args__", ()):
            self._register_imports(arg)

    def _translate_global_annotation(self, annotation, source_class_or_function):
        # convenience wrapper for _translate_annotation when the translated entity itself
        # determines eval scope and synchronizer target

        # infers synchronizer, target and home_module from an entity (class, function) containing the annotation
        synchronicity_target_interface = getattr(source_class_or_function, TARGET_INTERFACE_ATTR, None)
        synchronizer = getattr(source_class_or_function, SYNCHRONIZER_ATTR, None)
        if synchronizer:
            home_module = safe_get_module(getattr(source_class_or_function, synchronizer._original_attr))
        else:
            home_module = safe_get_module(source_class_or_function)

        return self._translate_annotation(annotation, synchronizer, synchronicity_target_interface, home_module)

    def _translate_annotation(
        self,
        annotation,
        synchronizer: typing.Optional[synchronicity.Synchronizer],
        synchronicity_target_interface: typing.Optional[Interface],
        home_module: typing.Optional[str],
    ):
        """
        Takes an annotation (type, generic, typevar, forward ref) and applies recursively (in case of generics):
        * eval for string annotations (importing `home_module` to be used as namespace)
        * re-mapping of the annotation to the correct synchronicity target
          (using synchronizer and synchronicity_target_interface)
        * registers imports for all referenced modules
        """
        if isinstance(annotation, typing.ForwardRef):  # TypeVars wrap their arguments as ForwardRefs (sometimes?)
            annotation = annotation.__forward_arg__
        if isinstance(annotation, str):
            try:
                annotation = evaluated_annotation(annotation, declaration_module=home_module)
            except Exception:
                logger.exception(
                    f"Error when evaluating {annotation} in {home_module}. Falling back to string annotation"
                )
                return annotation
        if type(annotation) is list:  # not using isinstance since ParamSpec is a subclass of list in Python 3.9
            return [
                self._translate_annotation(x, synchronizer, synchronicity_target_interface, home_module)
                for x in annotation
            ]

        translated_annotation = self._translate_annotation_map_types(
            annotation,
            synchronizer=synchronizer,
            interface=synchronicity_target_interface,
            home_module=home_module,
        )
        self._register_imports(translated_annotation)
        return translated_annotation

    def _translate_annotation_map_types(
        self,
        type_annotation,
        synchronizer: typing.Optional[synchronicity.Synchronizer],
        interface: typing.Optional[Interface],
        home_module: typing.Optional[str] = None,
    ):
        # recursively map a nested type annotation to match the output interface
        origin = getattr(type_annotation, "__origin__", None)
        args = safe_get_args(type_annotation)

        if isinstance(type_annotation, (typing_extensions.ParamSpecArgs, typing_extensions.ParamSpecKwargs)):
            # ParamSpecArgs and ParamSpecKwargs are special - they have an origin (the ParamSpec) but no attrs
            # we need to translate the origin in case it's a translated type annotation
            translated_origin = type_annotation.__origin__
            if synchronizer:
                translated_origin = synchronizer._translate_out(translated_origin, interface)
            return type(type_annotation)(translated_origin)

        elif origin is None or args is None:
            # TODO(elias): handle translation of un-parameterized async entities, like `Awaitable`
            # scalar - if type is synchronicity origin type, use the blocking/async version instead
            if synchronizer:
                return synchronizer._translate_out(type_annotation, interface)
            return type_annotation

        # Generics
        if origin == typing.Literal:
            mapped_args = args
        else:
            mapped_args = tuple(self._translate_annotation(arg, synchronizer, interface, home_module) for arg in args)

        if interface == Interface.BLOCKING:
            # blocking interface special generic translations:
            if origin == collections.abc.AsyncGenerator:
                return typing.Generator[mapped_args + (None,)]  # type: ignore[valid-type,misc]

            if origin == contextlib.AbstractAsyncContextManager:
                # TODO: in Python 3.13 mapped_args has a second argument for the exit type of the context
                #  manager, but we ignore that for now
                return combined_types.AsyncAndBlockingContextManager[mapped_args[0]]  # type: ignore[valid-type]

            if origin == collections.abc.AsyncIterable:
                return typing.Iterable[mapped_args]  # type: ignore[valid-type]

            if origin == collections.abc.AsyncIterator:
                return typing.Iterator[mapped_args]  # type: ignore[valid-type]

            if origin == collections.abc.Awaitable:
                return mapped_args[0]

            if origin == collections.abc.Coroutine:
                return mapped_args[2]

        # first see if the generic itself needs translation (in case of wrapped custom generics)
        if safe_get_module(origin) not in (
            "typing",
            "collections.abc",
            "contextlib",
            "builtins",
        ):  # don't translate built in generics in type annotations, even if they have been synchronicity wrapped
            # for base-class compatibility (e.g. AsyncContextManager, typing.Generic), otherwise it will break typing
            translated_origin = self._translate_annotation(origin, synchronizer, interface, home_module)
            t = translated_origin[mapped_args]  # type: ignore  # this seems to fall back to the __class_getitem__ of the implementation class
            # In order to get the right origin and args on the output, we manuall have to assign them:
            # TODO: We could probably fix this in the synchronicity layer by making wrapped generics true generics, or
            #  hackily by not letting __class_getitem__ proxy to the wrapped class' method for custom generics
            t.__module__ = safe_get_module(translated_origin)
            t.__origin__ = translated_origin
            t.__args__ = mapped_args
            return t

        return generic_copy_with_args(type_annotation, mapped_args)

    def _custom_signature(self, func, transform_signature=None) -> str:
        """
        We use this instead o str(inspect.Signature()) due to a few issues:
        * Generics with None args are incorrectly encoded as NoneType in str(signature)
        * Some names for stdlib module object types omit the module qualification (notably typing)
        * We might have to stringify annotations to support forward/self references
        * General flexibility like not being able to maintain *comments* in the arg declarations if we want to
        * We intentionally do not use follow_wrapped,
          since it will override runtime-transformed annotations on a wrapper
        * TypeVars default repr is `~T` instead of `origin_module.T` etc.
        """

        interface = getattr(func, TARGET_INTERFACE_ATTR, None)
        synchronizer = getattr(func, SYNCHRONIZER_ATTR, None)

        if synchronizer:
            home_module = safe_get_module(getattr(func, synchronizer._original_attr))
        else:
            home_module = safe_get_module(func)

        root_func = func
        if interface and synchronizer:
            root_func = synchronizer._translate_in(func)

        sig = sigtools.specifiers.signature(root_func)

        if sig.upgraded_return_annotation is not EmptyAnnotation:
            raw_return_annotation = sig.upgraded_return_annotation.source_value()
            return_annotation = self._translate_annotation(raw_return_annotation, synchronizer, interface, home_module)
            sig = sig.replace(
                return_annotation=return_annotation,
                upgraded_return_annotation=UpgradedAnnotation.upgrade(
                    return_annotation, func, None
                ),  # not sure if needed
            )

        new_parameters = []
        for param in sig.parameters.values():
            if param.upgraded_annotation is not EmptyAnnotation:
                raw_annotation = param.upgraded_annotation.source_value()
                translated_annotation = self._translate_annotation(raw_annotation, synchronizer, interface, home_module)
            elif param.annotation != inspect._empty:
                raw_annotation = param.annotation
                translated_annotation = self._translate_annotation(raw_annotation, synchronizer, interface, home_module)
            else:
                translated_annotation = param.annotation

            new_parameters.append(
                param.replace(
                    annotation=translated_annotation,
                    upgraded_annotation=UpgradedAnnotation.upgrade(
                        translated_annotation, func, param.name
                    ),  # not sure if needed...
                )
            )

        sig = sig.replace(parameters=new_parameters)

        if transform_signature:
            sig = transform_signature(sig)
        # kind of ugly, but this ensures valid formatting of Generics etc, see docstring above and _formatannotation
        with mock.patch("inspect.formatannotation", self._formatannotation):
            return str(sig)

    def _get_var_annotation(self, name, annotation):
        # TODO: how to translate annotation here - we don't know the home module
        self._register_imports(annotation)
        return f"{name}: {self._formatannotation(annotation, None)}"

    def _formatannotation(self, annotation, base_module=None) -> str:
        """modified version of `inspect.formatannotations`
        * Uses verbatim `None` instead of `NoneType` for None-arguments in generic types
        * Doesn't omit `typing.`-module from qualified imports in type names
        * ignores base_module (uses self.target_module instead)
        """
        origin = getattr(annotation, "__origin__", None)

        # For forward refs from modules that cannot be evaluated at runtime
        # but can be used freely in stub files, import the module and return the
        # original annotation
        if isinstance(annotation, typing.ForwardRef):
            if hasattr(annotation, "__forward_module__") and annotation.__forward_module__ in TYPE_CHECKING_OVERRIDES:
                self.imports.add(annotation.__forward_module__)
                return annotation.__forward_arg__

        # The remaining forward refs should already have been evaluated
        assert not isinstance(annotation, typing.ForwardRef)
        args = safe_get_args(annotation)

        if isinstance(annotation, typing_extensions.ParamSpecArgs):
            return self._formatannotation(typing_extensions.get_origin(annotation)) + "." + "args"

        if isinstance(annotation, typing_extensions.ParamSpecKwargs):
            return self._formatannotation(typing_extensions.get_origin(annotation)) + "." + "kwargs"

        if origin is None or not args:
            if annotation == typing.Sized:
                return "typing.Sized"  # fix Python 3.8(+?) where the repr is "typing.Sized[]" for some reason
            if annotation == typing.Hashable:
                return "typing.Hashable"  # fix Python 3.8(+?) where the repr is "typing.Hashable[]" for some reason

            if annotation == Ellipsis:
                return "..."
            if isinstance(annotation, type) or isinstance(annotation, (TypeVar, typing_extensions.ParamSpec)):
                if annotation is type(None):  # check for "NoneType"
                    return "None"
                name = (
                    annotation.__qualname__  # type: ignore
                    if hasattr(annotation, "__qualname__")
                    else annotation.__name__
                )
                annotation_module = safe_get_module(annotation)
                if annotation_module in ("builtins", self.target_module):
                    return name
                if annotation_module is None:
                    raise Exception(
                        f"{annotation} has __module__ == None - did you forget"
                        " to specify target module on a blocking type?"
                    )
                return annotation_module + "." + name
            if isinstance(annotation, list):
                # e.g. first argument to typing.Callable
                subargs = ",".join([self._formatannotation(arg) for arg in annotation])
                return f"[{subargs}]"

            return repr(annotation)

        # generic:
        origin_name = get_specific_generic_name(annotation)
        if origin is contextlib.AbstractAsyncContextManager:
            # python 3.13 adds a second optional exit arg, we only want to emit the first one
            # to be backwards compatible
            args = args[:1]

        if (safe_get_module(annotation), origin_name) == ("typing", "Optional"):
            # typing.Optional adds a None argument that we shouldn't include when formatting
            (optional_arg,) = [a for a in args if a is not type(None)]
            comma_separated_args = self._formatannotation(optional_arg)
        else:
            formatted_args = [self._formatannotation(a) for a in args]
            comma_separated_args = ", ".join(formatted_args)

        annotation_module = safe_get_module(annotation)
        if annotation_module in ("typing", "contextlib") and origin_name.startswith("Abstract"):
            # This is needed for Python <=3.8 where there is a bug (?) in the typing.AsyncContextManager
            # causing it to be represented with the non-existent name typing.AbstractContextManager
            # >>> typing.AsyncContextManager
            # typing.AbstractAsyncContextManager
            origin_name = origin_name[len("Abstract") :]  # cut the "Abstract"

        if annotation_module not in ("builtins", self.target_module):
            # need to qualify the module of the origin
            origin_module = annotation_module
            origin_name = f"{origin_module}.{origin_name}"

        return f"{origin_name}[{comma_separated_args}]"

    def _indent(self, level):
        return level * self._indentation

    def _get_function_source_with_overloads(self, func, name, indentation_level=0, transform_signature=None) -> str:
        signature_indent = self._indent(indentation_level)
        body_indent = self._indent(indentation_level + 1)
        parts = []

        interface = func.__dict__.get(TARGET_INTERFACE_ATTR)

        synchronizer = func.__dict__.get(SYNCHRONIZER_ATTR)
        if interface:
            root_func = func.__dict__[SYNCHRONIZER_ATTR]._translate_in(func)
        else:
            root_func = func

        overloaded_signatures = overload_tracking.get_overloads(root_func)
        for overload_func in overloaded_signatures:
            self.imports.add("typing")
            parts.append(f"{signature_indent}@typing.overload")

            if interface:
                overload_func = synchronizer._wrap(overload_func, interface, name=name)

            overload_src = self._get_function_source(
                overload_func,
                name,
                signature_indent,
                body_indent,
                transform_signature=transform_signature,
            )
            parts.append(overload_src)

        if not overloaded_signatures:
            # only add the functions complete signatures if there are no stubs
            parts.append(
                self._get_function_source(
                    func,
                    name,
                    signature_indent,
                    body_indent,
                    transform_signature=transform_signature,
                )
            )
        return "\n".join(parts)

    def _get_function_source(
        self,
        func,
        name,
        signature_indent: str,
        body_indent: str,
        transform_signature=None,
    ) -> str:
        maybe_decorators = ""
        if hasattr(func, "__dataclass_transform__"):
            dt_spec = func.__dataclass_transform__
            if dt_spec["field_specifiers"]:
                refs = ""
                for field_spec_entity in dt_spec["field_specifiers"]:
                    field_spec_module = safe_get_module(field_spec_entity)
                    if field_spec_module == self.target_module:
                        ref = field_spec_entity.__qualname__
                    else:
                        self.imports.add(field_spec_module)
                        ref = f"{field_spec_module}.{field_spec_entity.__qualname__}"
                    refs += ref + ", "

                args = f"field_specifiers=({refs}), "
            bool_attrs = {"eq_default": True, "order_default": False, "kw_only_default": False, "frozen_default": False}
            for k, v in bool_attrs.items():
                if dt_spec[k] != v:
                    args += f"{k}={dt_spec[k]}, "

            self.imports.add("typing_extensions")
            maybe_decorators = f"{signature_indent}@typing_extensions.dataclass_transform({args})\n"

        async_prefix = ""
        if is_coroutine_function_follow_wrapped(func):
            # note: async prefix should not be used for annotated abstract/stub *async generators*,
            # so we don't check for inspect.isasyncgenfunction since they contain no yield keyword,
            # and would otherwise indicate an awaitable that returns an async generator to static type checkers
            async_prefix = "async "

        signature = self._custom_signature(func, transform_signature)

        return "\n".join(
            [
                f"{maybe_decorators}{signature_indent}{async_prefix}def {name}{signature}:",
                f"{body_indent}...",
                "",
            ]
        )


def write_stub(module_path: str):
    with overload_tracking.patched_overload():
        mod = importlib.import_module(module_path)
        assert mod.__file__ is not None
        emitter = StubEmitter.from_module(mod)
        source = emitter.get_source()
        stub_path = Path(mod.__file__).with_suffix(".pyi")
        stub_path.write_text(source)
        return stub_path


if __name__ == "__main__":
    for module_path in sys.argv[1:]:
        out_path = write_stub(module_path)
        print(f"Wrote {out_path}")



# this code is only meant to be "running" through mypy and not an actual python interpreter!
import typing

from typing_extensions import assert_type

from test.type_stub_helpers import e2e_example_export

blocking_foo = e2e_example_export.BlockingFoo("hello")

# assert start
assert_type(blocking_foo, e2e_example_export.BlockingFoo)

assert_type(blocking_foo.getarg(), str)
assert_type(blocking_foo.gen(), typing.Generator[int, None, None])

assert_type(e2e_example_export.some_instance, typing.Optional[e2e_example_export.BlockingFoo])

assert_type(blocking_foo.some_static("foo"), float)

assert_type(e2e_example_export.BlockingFoo.clone(blocking_foo), e2e_example_export.BlockingFoo)

assert_type(blocking_foo.singleton, e2e_example_export.BlockingFoo)


assert_type(
    e2e_example_export.listify(blocking_foo),
    typing.List[e2e_example_export.BlockingFoo],
)


assert_type(e2e_example_export.overloaded("12"), float)

assert_type(e2e_example_export.overloaded(12), int)


with e2e_example_export.wrapped_make_context(10.0) as c:
    assert_type(c, str)


async def async_block() -> None:
    res = await e2e_example_export.returns_foo.aio()
    assert_type(res, e2e_example_export.BlockingFoo)

    async with e2e_example_export.wrapped_make_context(10.0) as c:
        assert_type(c, str)

    # not sure if this should actually be supported, but it is, for completeness:
    async with e2e_example_export.wrapped_make_context.aio(10.0) as c:
        assert_type(c, str)


def f(a: str) -> float:
    return 0.1


res = e2e_example_export.wrap_callable(f).func(a="q")
assert_type(res, float)



import asyncio
import signal
import sys
import threading
import typing

from synchronicity.exceptions import NestedEventLoops

T = typing.TypeVar("T")


class Runner:
    """Simplified backport of asyncio.Runner from Python 3.11

    Like asyncio.run() but allows multiple calls to the same event loop
    before teardown, and is converts sigints into graceful cancellations
    similar to asyncio.run on Python 3.11+.
    """

    def __enter__(self) -> "Runner":
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            pass  # no event loop - this is what we expect!
        else:
            raise NestedEventLoops()

        self._loop = asyncio.new_event_loop()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._loop.run_until_complete(self._loop.shutdown_asyncgens())
        if sys.version_info[:2] >= (3, 9):
            # Introduced in Python 3.9
            self._loop.run_until_complete(self._loop.shutdown_default_executor())

        self._loop.close()
        return False

    def run(self, coro: typing.Awaitable[T]) -> T:
        is_main_thread = threading.current_thread() == threading.main_thread()
        self._num_sigints = 0

        coro_task = asyncio.ensure_future(coro, loop=self._loop)

        async def wrapper_coro():
            # this wrapper ensures that we won't reraise KeyboardInterrupt into
            # the calling scope until all async finalizers in coro_task have
            # finished executing. It even allows the coro to prevent cancellation
            # and thereby ignoring the first keyboardinterrupt
            return await coro_task

        def _sigint_handler(signum, frame):
            # cancel the task in order to have run_until_complete return soon and
            # prevent a bunch of unwanted tracebacks when shutting down the
            # event loop.

            # this basically replicates the sigint handler installed by asyncio.run()
            self._num_sigints += 1
            if self._num_sigints == 1:
                # first sigint is graceful
                self._loop.call_soon_threadsafe(coro_task.cancel)
                return

            # this should normally not happen, but the second sigint would "hard kill" the event loop
            # by raising KeyboardInterrupt inside of it
            raise KeyboardInterrupt()

        original_sigint_handler = None
        try:
            # only install signal handler if running from main thread and we haven't disabled sigint
            handle_sigint = is_main_thread and signal.getsignal(signal.SIGINT) == signal.default_int_handler

            if handle_sigint:
                # intentionally not using _loop.add_signal_handler since it's slow (?)
                # and not available on Windows. We just don't want the sigint to
                # mess with the event loop anyways
                original_sigint_handler = signal.signal(signal.SIGINT, _sigint_handler)
        except KeyboardInterrupt:
            # this is quite unlikely, but with bad timing we could get interrupted before
            # installing the sigint handler and this has happened repeatedly in unit tests
            _sigint_handler(signal.SIGINT, None)

        try:
            return self._loop.run_until_complete(wrapper_coro())
        except asyncio.CancelledError:
            if self._num_sigints > 0:
                raise KeyboardInterrupt()  # might want to use original_sigint_handler here instead?
            raise  # "internal" cancellations, not triggered by KeyboardInterrupt
        finally:
            if original_sigint_handler:
                # reset signal handler
                signal.signal(signal.SIGINT, original_sigint_handler)



import collections
import functools
import importlib
import pytest
import sys
import typing
from textwrap import dedent

import typing_extensions

import synchronicity
from synchronicity import overload_tracking
from synchronicity.async_wrap import asynccontextmanager
from synchronicity.type_stubs import StubEmitter

from .type_stub_helpers import some_mod


def noop(): ...


def arg_no_anno(arg1): ...


def scalar_args(arg1: str, arg2: int) -> float:
    return 0


def generic_other_module_arg(arg: typing.List[some_mod.Foo]): ...


async def async_func() -> str:
    return "hello"


def _function_source(func, target_module=__name__):
    stub_emitter = StubEmitter(target_module)
    stub_emitter.add_function(func, func.__name__)
    return stub_emitter.get_source()


def _class_source(cls, target_module=__name__):
    stub_emitter = StubEmitter(target_module)
    stub_emitter.add_class(cls, cls.__name__)
    return stub_emitter.get_source()


def test_function_basics():
    assert _function_source(noop) == "def noop():\n    ...\n"
    assert _function_source(arg_no_anno) == "def arg_no_anno(arg1):\n    ...\n"
    assert _function_source(scalar_args) == "def scalar_args(arg1: str, arg2: int) -> float:\n    ...\n"


def test_function_with_imports():
    assert (
        _function_source(generic_other_module_arg, target_module="dummy")
        == """import test.type_stub_helpers.some_mod
import typing

def generic_other_module_arg(arg: typing.List[test.type_stub_helpers.some_mod.Foo]):
    ...
"""
    )


def test_async_func():
    assert _function_source(async_func) == "async def async_func() -> str:\n    ...\n"


def test_async_gen():
    async def async_gen() -> typing.AsyncGenerator[int, None]:
        yield 0

    assert (
        _function_source(async_gen)
        == "import typing\n\ndef async_gen() -> typing.AsyncGenerator[int, None]:\n    ...\n"
    )

    def weird_async_gen() -> typing.AsyncGenerator[int, None]:
        # non-async function that returns an async generator
        async def gen():
            yield 0

        return gen()

    assert (
        _function_source(weird_async_gen)
        == "import typing\n\ndef weird_async_gen() -> typing.AsyncGenerator[int, None]:\n    ...\n"
    )

    async def it() -> typing.AsyncIterator[str]:  # this is the/a correct annotation
        yield "hello"

    src = _function_source(it)
    assert "yield" not in src
    # since the yield keyword is removed in a type stub, the async prefix needs to be removed as well
    # to avoid "double asyncness" (while keeping the remaining annotation)
    assert "async" not in src
    assert "def it() -> typing.AsyncIterator[str]:" in src


class MixedClass:
    class_var: str

    def some_method(self) -> bool:
        return False

    @classmethod
    def some_class_method(cls) -> int:
        return 1

    @staticmethod
    def some_staticmethod() -> float:
        return 0.0

    @property
    def some_property(self) -> str:
        return ""

    @some_property.setter
    def some_property(self, val):
        print(val)

    @some_property.deleter
    def some_property(self, val):
        print(val)


def test_class_generation():
    emitter = StubEmitter(__name__)
    emitter.add_class(MixedClass, "MixedClass")
    source = emitter.get_source()
    last_assertion_location = None

    def assert_in_after_last(search_string):
        nonlocal last_assertion_location
        assert search_string in source
        if last_assertion_location is not None:
            new_location = source.find(search_string)
            assert new_location > last_assertion_location
            last_assertion_location = new_location

    indent = "    "
    assert_in_after_last("class MixedClass:")
    assert_in_after_last(f"{indent}class_var: str")
    assert_in_after_last(f"{indent}class_var: str")
    assert_in_after_last(f"{indent}def some_method(self) -> bool:\n{indent * 2}...")
    assert_in_after_last(f"{indent}@classmethod\n{indent}def some_class_method(cls) -> int:\n{indent * 2}...")
    assert_in_after_last(f"{indent}@staticmethod\n{indent}def some_staticmethod() -> float:")
    assert_in_after_last(f"{indent}@property\n{indent}def some_property(self) -> str:")
    assert_in_after_last(f"{indent}@some_property.setter\n{indent}def some_property(self, val):")
    assert_in_after_last(f"{indent}@some_property.deleter\n{indent}def some_property(self, val):")


def merged_signature(*sigs):
    sig = sigs[0].copy()
    return sig


def test_wrapped_function_with_new_annotations():
    """A wrapped function (in general, using functools.wraps/partial) would
    have an inspect.signature from the wrapped function by default
    and from the wrapper function is inspect.signature gets the follow_wrapped=True
    option. However, for the best type stub usability, the best would be to combine
    all layers of wrapping, adding any additional arguments or annotations as updates
    to the underlying wrapped function signature.

    This test makes sure we do just that.
    """

    def orig(arg: str): ...

    @functools.wraps(orig)
    def wrapper(extra_arg: int, *args, **kwargs):
        orig(*args, **kwargs)

    wrapper.__annotations__.update({"extra_arg": int, "arg": float})
    assert _function_source(wrapper) == "def orig(extra_arg: int, arg: float):\n    ...\n"


def test_wrapped_async_func_remains_async():
    async def orig(arg: str): ...

    @functools.wraps(orig)
    def wrapper(*args, **kwargs):
        return orig(*args, **kwargs)

    assert _function_source(wrapper) == "async def orig(arg: str):\n    ...\n"


class Base:
    def base_method(self) -> str:
        return ""


Base.__module__ = "basemod"
Base.__qualname__ = "Base"


class Based(Base):
    def sub(self) -> float:
        return 0


def test_base_class_included_and_imported():
    src = _class_source(Based)
    assert "import basemod" in src
    assert "class Based(basemod.Base):" in src
    assert "base_method" not in src  # base class method should not be in emitted stub


def test_typevar():
    T = typing.TypeVar("T")
    T.__module__ = "source_mod"

    def foo(arg: T) -> T:
        return arg

    src = _function_source(foo)
    assert "import source_mod" in src
    assert "def foo(arg: source_mod.T) -> source_mod.T" in src


def test_string_annotation():
    stub_emitter = StubEmitter("dummy")
    stub_emitter.add_variable(annotation="Foo", name="some_foo")  # string annotation
    src = stub_emitter.get_source()
    assert 'some_foo: "Foo"' in src or "some_foo: 'Foo'" in src


class Forwarder:
    def foo(self) -> typing.List["Forwardee"]: ...


class Forwardee: ...


def test_forward_ref():
    # add in the same order here:
    stub = StubEmitter(__name__)
    stub.add_class(Forwarder, "Forwarder")
    stub.add_class(Forwardee, "Forwardee")
    src = stub.get_source()
    assert "class Forwarder:" in src
    assert (
        "def foo(self) -> typing.List[Forwardee]:" in src
    )  # should technically be quoted 'Forwardee', but non-strings seem ok in pure type stubs


def test_optional():
    # Not super important, but try to preserve typing.Optional as typing.Optional instead of typing.Union[None, ...]
    # This only works on Python 3.10+, since 3.9 and earlier do "eager" conversion when creating the type
    def f() -> typing.Optional[str]: ...

    wrapped_f = synchronizer.create_blocking(f, "wrapped_f", __name__)

    src = _function_source(wrapped_f)
    if sys.version_info[:2] >= (3, 10):
        assert "typing.Optional[str]" in src
    else:
        assert "typing.Union[str, None]" in src


class SelfRefFoo:
    def foo(self) -> "SelfRefFoo":
        return self


def test_self_ref():
    src = _class_source(SelfRefFoo)
    assert (
        "def foo(self) -> SelfRefFoo" in src
    )  # should technically be 'Foo' but non-strings seem ok in pure type stubs


class _Foo:
    @staticmethod
    async def clone(foo: "_Foo") -> "_Foo":
        return foo


synchronizer = synchronicity.Synchronizer()
Foo = synchronizer.create_blocking(_Foo, "Foo", __name__)


def test_synchronicity_type_translation():
    async def _get_foo(foo: _Foo) -> typing.AsyncContextManager[_Foo]:
        return foo

    get_foo = synchronizer.create_blocking(_get_foo, "get_foo", __name__)
    src = _function_source(get_foo)

    print(src)
    assert "class __get_foo_spec(typing_extensions.Protocol):" in src
    assert "    def __call__(self, foo: Foo) -> synchronicity.combined_types.AsyncAndBlockingContextManager[Foo]" in src
    # python 3.13 has an exit type generic argument, e.g. typing.AsyncContextManager[Foo, bool | None]
    # but we want the type stubs to work on older versions of python too (without conditionals everywhere):
    assert "    async def aio(self, foo: Foo) -> typing.AsyncContextManager[Foo]" in src
    assert "get_foo: __get_foo_spec"


def test_synchronicity_wrapped_class():
    src = _class_source(Foo)
    print(src)
    # assert "__init__" not in Foo
    assert "class __clone_spec(typing_extensions.Protocol):" in src
    assert "    def __call__(self, foo: Foo) -> Foo" in src
    assert "    async def aio(self, foo: Foo) -> Foo" in src
    assert "clone: __clone_spec" in src


class _WithClassMethod:
    @classmethod
    def classy(cls): ...

    async def meth(self, arg: bool) -> int:
        return 0


WithClassMethod = synchronizer.create_blocking(_WithClassMethod, "WithClassMethod", __name__)


def test_synchronicity_class():
    src = _class_source(WithClassMethod)
    assert "    @classmethod" in src
    assert "    def classy(cls):" in src

    assert "__meth_spec" in src

    assert (
        """
    class __meth_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, arg: bool) -> int:
            ...

        async def aio(self, arg: bool) -> int:
            ...

    meth: __meth_spec[typing_extensions.Self]
"""
        in src
    )


T = typing.TypeVar("T")
P = typing_extensions.ParamSpec("P")


Translated_T = synchronizer.create_blocking(T, "Translated_T", __name__)
Translated_P = synchronizer.create_blocking(P, "Translated_P", __name__)


class MyGeneric(typing.Generic[T]): ...


BlockingMyGeneric = synchronizer.create_blocking(
    MyGeneric,
    "BlockingMyGeneric",
    __name__,
)


def test_custom_generic():
    # TODO: build out this test a bit, as it currently creates an invalid stub (missing base types)
    src = _class_source(BlockingMyGeneric)

    class Specific(MyGeneric[str]): ...

    src = _class_source(Specific)
    assert "class Specific(MyGeneric[str]):" in src


class ParamSpecGeneric(typing.Generic[P, T]):
    async def meth(self, *args: P.args, **kwargs: P.kwargs) -> typing_extensions.Self: ...

    def syncfunc(self) -> T: ...


BlockingParamSpecGeneric = synchronizer.create_blocking(ParamSpecGeneric, "BlockingParamSpecGeneric", __name__)


def test_paramspec_generic():
    src = _class_source(BlockingParamSpecGeneric)
    assert "class BlockingParamSpecGeneric(typing.Generic[Translated_P, Translated_T])" in src

    assert "class __meth_spec(typing_extensions.Protocol[Translated_P_INNER, SUPERSELF]):" in src
    assert "def __call__(self, *args: Translated_P_INNER.args, **kwargs: Translated_P_INNER.kwargs) -> SUPERSELF" in src
    assert "def aio(self, *args: Translated_P_INNER.args, **kwargs: Translated_P_INNER.kwargs) -> SUPERSELF" in src
    assert "meth: __meth_spec[Translated_P, typing_extensions.Self]" in src
    assert "def syncfunc(self) -> Translated_T:" in src


def test_synchronicity_generic_subclass():
    class Specific(MyGeneric[str]): ...

    assert Specific.__bases__ == (MyGeneric,)
    assert Specific.__orig_bases__ == (MyGeneric[str],)

    BlockingSpecific = synchronizer.create_blocking(Specific, "BlockingSpecific", __name__)
    assert BlockingSpecific.__bases__ == (BlockingMyGeneric,)
    assert BlockingSpecific.__orig_bases__ == (BlockingMyGeneric[str],)

    src = _class_source(BlockingSpecific)
    assert "class BlockingSpecific(BlockingMyGeneric[str]):" in src

    async def foo_impl(bar: MyGeneric[str]): ...

    foo = synchronizer.create_blocking(foo_impl, "foo")
    src = _function_source(foo)
    assert "def __call__(self, bar: BlockingMyGeneric[str]):" in src
    assert "async def aio(self, bar: BlockingMyGeneric[str]):" in src


_B = typing.TypeVar("_B", bound="str")

B = synchronizer.create_blocking(
    _B, "B", __name__
)  # only strictly needed if the bound is a synchronicity implementation type


def _ident(b: _B) -> _B:
    return b


ident = synchronizer.create_blocking(_ident, "ident", __name__)


def test_translated_bound_type_vars():
    emitter = StubEmitter(__name__)
    emitter.add_type_var(B, "B")
    emitter.add_function(ident, "ident")
    src = emitter.get_source()
    assert 'B = typing.TypeVar("B", bound="str")' in src
    assert "def ident(b: B) -> B" in src


def test_literal_alias(tmp_path):
    contents = dedent(
        """
        import typing
        from typing import Literal
        foo = typing.Literal["foo"]
        bar = Literal["bar"]
        """
    )
    with open(fname := (tmp_path / "foo.py"), "w") as f:
        f.write(contents)

    spec = importlib.util.spec_from_file_location("foo", fname)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    emitter = StubEmitter.from_module(mod)
    src = emitter.get_source()
    assert "foo = typing.Literal['foo']" in src
    assert "bar = typing.Literal['bar']" in src


def test_callable():
    def foo() -> collections.abc.Callable[[str], float]:
        return lambda x: 0.0

    src = _function_source(foo)
    assert "-> collections.abc.Callable[[str], float]" in src


def test_ellipsis():
    def foo() -> collections.abc.Callable[..., typing.Any]:
        return lambda x: 0

    src = _function_source(foo)
    assert "-> collections.abc.Callable[..., typing.Any]" in src


def test_param_spec():
    P = typing_extensions.ParamSpec("P")

    def foo() -> collections.abc.Callable[P, typing.Any]:
        return lambda x: 0

    src = _function_source(foo)
    assert "-> collections.abc.Callable[P, typing.Any]" in src


def test_typing_literal():
    def foo() -> typing.Literal["three", "str"]:
        return "str"

    src = _function_source(foo)
    assert "-> typing.Literal['three', 'str']" in src  # "str" should not be eval:ed in a Literal!


def test_overloads_unwrapped_functions():
    with overload_tracking.patched_overload():

        @typing.overload
        def _overloaded(arg: str) -> float: ...

        @typing.overload
        def _overloaded(arg: int) -> int: ...

        def _overloaded(arg: typing.Union[str, int]):
            if isinstance(arg, str):
                return float(arg)
            return arg

    overloaded = synchronizer.create_blocking(_overloaded, "overloaded")

    src = _function_source(overloaded)
    assert src.count("@typing.overload") == 2
    assert src.count("def overloaded") == 2  # original should be omitted
    assert "def overloaded(arg: str) -> float" in src
    assert "def overloaded(arg: int) -> int:" in src


def test_wrapped_context_manager_is_both_blocking_and_async():
    @asynccontextmanager
    async def foo(arg: int) -> typing.AsyncGenerator[str, None]:
        yield "hello"

    wrapped_foo = synchronizer.create_blocking(foo, name="wrapped_foo")
    assert wrapped_foo.__annotations__["return"] == typing.AsyncContextManager[str]
    wrapped_foo_src = _function_source(wrapped_foo)

    assert (
        "def __call__(self, arg: int) -> synchronicity.combined_types.AsyncAndBlockingContextManager[str]:"
        in wrapped_foo_src
    )
    assert "AbstractAsyncContextManager" not in wrapped_foo_src


@pytest.mark.skipif(sys.version_info < (3, 9), reason="collections.abc.Iterator isn't a generic type before Python 3.9")
def test_collections_iterator():
    def foo() -> collections.abc.Iterator[int]:
        class MyIterator(collections.abc.Iterator):
            def __iter__(self) -> collections.abc.Iterator[int]:
                return self

            def __next__(self) -> int:
                return 1

        return MyIterator()

    src = _function_source(foo)
    assert "-> collections.abc.Iterator[int]" in src


U = typing.TypeVar("U")


class _ReturnVal(typing.Generic[U]):
    pass


ReturnVal = synchronizer.create_blocking(_ReturnVal, "ReturnVal", __name__)


def test_returns_forward_wrapped_generic():
    # forward reference of a wrapped generic as a string is one of the trickier cases to handle
    # as the string needs to be evaluated, the generics need to be recursively expanded and
    # type vars need to be replaced with "inner" generated versions

    class _Container(typing.Generic[T]):
        async def fun(self) -> "ReturnVal[T]":
            return ReturnVal()

    Container = synchronizer.create_blocking(_Container, "Container")

    src = _class_source(Container)

    # base class should be generic in the (potentially) translated type var (could have wrapped bounds spec)
    assert "class Container(typing.Generic[Translated_T]):" in src
    assert "Translated_T_INNER = typing.TypeVar" in src  # distinct "inner copy" of Translated_T needs to be declared
    assert "typing_extensions.Protocol[Translated_T_INNER, SUPERSELF]" in src
    assert "def __call__(self) -> ReturnVal[Translated_T_INNER]:" in src
    assert "fun: __fun_spec[Translated_T, typing_extensions.Self]" in src


def custom_field():  # needs to be in global scope
    pass


def test_dataclass_transform():
    @typing_extensions.dataclass_transform(field_specifiers=(custom_field,), kw_only_default=True)
    def decorator():
        pass

    src = _function_source(decorator)
    assert "@typing_extensions.dataclass_transform(field_specifiers=(custom_field, ), kw_only_default=True, )\n" in src

    src = _function_source(decorator, target_module="other_module")
    assert "import test.type_stub_test" in src
    assert "import typing_extensions" in src
    assert "field_specifiers=(test.type_stub_test.custom_field, )" in src


def test_contextvar():
    import contextvars

    s = StubEmitter("blah")
    s.add_variable(contextvars.ContextVar, "c")
    src = s.get_source()
    assert "import contextvars" in src
    assert "c: contextvars.ContextVar" in src


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="collections.abc.Callable strips Concatenate wrappers at runtime before Python 3.10 :(",
)
def test_concatenate_origin_module():
    s = StubEmitter(__name__)
    P = typing_extensions.ParamSpec("P")
    R = typing.TypeVar("R")
    s.add_variable(collections.abc.Callable[typing_extensions.Concatenate[typing.Any, P], R], "f")
    src = s.get_source()
    print(src)
    assert "f: collections.abc.Callable[typing_extensions.Concatenate[typing.Any, P], R]" in src


def test_paramspec_args():
    from .type_stub_helpers.some_mod import P

    def foo(fn: collections.abc.Callable[P, None], *args: P.args, **kwargs: P.kwargs) -> str:
        return "Hello World!"

    src = _function_source(foo)
    assert "import test.type_stub_helpers.some_mod" in src
    assert (
        "def foo(fn: collections.abc.Callable[test.type_stub_helpers.some_mod.P, None], *args: test.type_stub_helpers.some_mod.P.args, **kwargs: test.type_stub_helpers.some_mod.P.kwargs) -> str:"  # noqa
        in src
    )  # noqa: E501


if typing.TYPE_CHECKING:
    import _typeshed


def test_typeshed():
    """Test that _typeshed annotations are preserved in stubs."""

    def foo() -> "_typeshed.OpenTextMode":
        return "r"

    src = _function_source(foo)
    assert "import _typeshed" in src
    assert "def foo() -> _typeshed.OpenTextMode:" in src



import pytest
import subprocess
import sys
from pathlib import Path


@pytest.mark.skipif(sys.platform == "win32", reason="Windows can't fork")
def test_fork_restarts_loop():
    p = subprocess.Popen(
        [sys.executable, Path(__file__).parent / "support" / "_forker.py"],
        encoding="utf8",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    try:
        stdout, stderr = p.communicate(timeout=2)
    except subprocess.TimeoutExpired:
        p.kill()
        assert False, "Fork process hanged"

    assert p.returncode == 0
    assert stdout == "done\ndone\n"



import asyncio
import concurrent.futures
import inspect
import pytest
import time
import typing
from typing import Coroutine
from unittest.mock import MagicMock

import synchronicity
from synchronicity import Synchronizer

SLEEP_DELAY = 0.5


async def f(x):
    await asyncio.sleep(SLEEP_DELAY)
    return x**2


async def f2(fn, x):
    return await fn(x)


def test_function_sync(synchronizer):
    s = synchronizer
    t0 = time.time()
    f_s = s.create_blocking(f)
    assert f_s.__name__ == "blocking_f"
    ret = f_s(42)
    assert ret == 1764
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


def test_function_sync_future(synchronizer):
    t0 = time.time()
    f_s = synchronizer.create_blocking(f)
    assert f_s.__name__ == "blocking_f"
    fut = f_s(42, _future=True)
    assert isinstance(fut, concurrent.futures.Future)
    assert time.time() - t0 < SLEEP_DELAY
    assert fut.result() == 1764
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


@pytest.mark.asyncio
async def test_function_async_as_function_attribute(synchronizer):
    s = synchronizer
    t0 = time.time()
    f_s = s.create_blocking(f).aio
    assert f_s.__name__ == "aio_f"
    coro = f_s(42)
    assert inspect.iscoroutine(coro)
    assert time.time() - t0 < SLEEP_DELAY
    assert await coro == 1764
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY

    # Make sure the same-loop calls work
    f2_s = s.create_blocking(f2).aio
    assert f2_s.__name__ == "aio_f2"
    coro = f2_s(f_s, 42)
    assert await coro == 1764

    # Make sure cross-loop calls work
    s2 = Synchronizer()
    f2_s2 = s2.create_blocking(f2).aio
    assert f2_s2.__name__ == "aio_f2"
    coro = f2_s2(f_s, 42)
    assert await coro == 1764
    s2._close_loop()


@pytest.mark.asyncio
async def test_function_async_block_event_loop(synchronizer):
    async def spinlock():
        # This blocks the event loop, but not the main event loop
        time.sleep(SLEEP_DELAY)

    spinlock_s = synchronizer.create_blocking(spinlock)
    spinlock_coro = spinlock_s.aio()
    sleep_coro = asyncio.sleep(SLEEP_DELAY)

    t0 = time.time()
    await asyncio.gather(spinlock_coro, sleep_coro)
    assert SLEEP_DELAY <= time.time() - t0 < 2 * SLEEP_DELAY


def test_function_many_parallel_sync(synchronizer):
    g = synchronizer.create_blocking(f)
    t0 = time.time()
    rets = [g(i) for i in range(10)]  # Will resolve serially
    assert len(rets) * SLEEP_DELAY < time.time() - t0 < (len(rets) + 1) * SLEEP_DELAY


def test_function_many_parallel_sync_futures(synchronizer):
    g = synchronizer.create_blocking(f)
    t0 = time.time()
    futs = [g(i, _future=True) for i in range(100)]
    assert isinstance(futs[0], concurrent.futures.Future)
    assert time.time() - t0 < SLEEP_DELAY
    assert [fut.result() for fut in futs] == [z**2 for z in range(100)]
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


@pytest.mark.asyncio
async def test_function_many_parallel_async(synchronizer):
    g = synchronizer.create_blocking(f)
    t0 = time.time()
    coros = [g.aio(i) for i in range(100)]
    assert inspect.iscoroutine(coros[0])
    assert time.time() - t0 < SLEEP_DELAY
    assert await asyncio.gather(*coros) == [z**2 for z in range(100)]
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


async def gen(n):
    for i in range(n):
        await asyncio.sleep(SLEEP_DELAY)
        yield i


async def gen2(generator, n):
    async for ret in generator(n):
        yield ret


def test_generator_sync(synchronizer):
    synchronizer = synchronizer
    t0 = time.time()
    gen_s = synchronizer.create_blocking(gen)
    it = gen_s(3)
    assert inspect.isgenerator(it)
    assert time.time() - t0 < SLEEP_DELAY
    lst = list(it)
    assert lst == [0, 1, 2]
    assert time.time() - t0 > len(lst) * SLEEP_DELAY


@pytest.mark.asyncio
async def test_generator_async(synchronizer):
    t0 = time.time()
    gen_s = synchronizer.create_blocking(gen).aio

    asyncgen = gen_s(3)
    assert inspect.isasyncgen(asyncgen)
    assert time.time() - t0 < SLEEP_DELAY
    lst = [z async for z in asyncgen]
    assert lst == [0, 1, 2]
    assert time.time() - t0 > len(lst) * SLEEP_DELAY

    # Make sure same-loop calls work
    gen2_s = synchronizer.create_blocking(gen2).aio
    asyncgen = gen2_s(gen_s, 3)
    lst = [z async for z in asyncgen]
    assert lst == [0, 1, 2]

    # Make sure cross-loop calls work
    s2 = synchronizer
    gen2_s2 = s2.create_blocking(gen2).aio
    asyncgen = gen2_s2(gen_s, 3)
    lst = [z async for z in asyncgen]
    assert lst == [0, 1, 2]


@pytest.mark.asyncio
async def test_function_returning_coroutine(synchronizer):
    def func() -> Coroutine:
        async def inner():
            return 10

        return inner()

    blocking_func = synchronizer.create_blocking(func)
    assert blocking_func() == 10
    coro = blocking_func.aio()
    assert inspect.iscoroutine(coro)
    assert await coro == 10


def test_sync_lambda_returning_coroutine_sync(synchronizer):
    t0 = time.time()
    g = synchronizer.create_blocking(lambda z: f(z + 1))
    ret = g(42)
    assert ret == 1849
    assert time.time() - t0 >= SLEEP_DELAY


def test_sync_lambda_returning_coroutine_sync_futures(synchronizer):
    t0 = time.time()
    g = synchronizer.create_blocking(lambda z: f(z + 1))
    fut = g(42, _future=True)
    assert isinstance(fut, concurrent.futures.Future)
    assert time.time() - t0 < SLEEP_DELAY
    assert fut.result() == 1849
    assert time.time() - t0 >= SLEEP_DELAY


@pytest.mark.asyncio
async def test_sync_inline_func_returning_coroutine_async(synchronizer):
    t0 = time.time()

    # NOTE: we don't create the async variant unless we know the function returns a coroutine
    def func(z) -> Coroutine:
        return f(z + 1)

    g = synchronizer.create_blocking(func)
    coro = g.aio(42)
    assert inspect.iscoroutine(coro)
    assert time.time() - t0 < SLEEP_DELAY
    assert await coro == 1849
    assert time.time() - t0 >= SLEEP_DELAY


class Base:
    def __init__(self, x):
        self._x = x


class MyClass(Base):
    def __init__(self, x):
        super().__init__(x)

    async def start(self):
        async def task():
            await asyncio.sleep(SLEEP_DELAY)
            return self._x

        loop = asyncio.get_event_loop()
        self._task = loop.create_task(task())

    async def get_result(self):
        ret = await self._task
        return ret**2

    async def __aenter__(self):
        await asyncio.sleep(SLEEP_DELAY)
        return 42

    async def __aexit__(self, exc_type, exc, tb):
        await asyncio.sleep(SLEEP_DELAY)

    @staticmethod
    async def my_static_method():
        await asyncio.sleep(SLEEP_DELAY)
        return 43

    @classmethod
    async def my_class_method(cls):
        await asyncio.sleep(SLEEP_DELAY)
        return 44

    async def __aiter__(self):
        for i in range(self._x):
            yield i


def test_class_sync(synchronizer):
    BlockingBase = synchronizer.create_blocking(Base, name="BlockingBase")
    BlockingMyClass = synchronizer.create_blocking(MyClass, name="BlockingMyClass")

    assert BlockingMyClass.__name__ == "BlockingMyClass"
    obj = BlockingMyClass(x=42)
    assert isinstance(obj, BlockingMyClass)
    assert isinstance(obj, BlockingBase)
    obj.start()
    ret = obj.get_result()
    assert ret == 1764

    t0 = time.time()
    with obj as z:
        assert z == 42
        assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY
    assert time.time() - t0 > 2 * SLEEP_DELAY

    t0 = time.time()
    assert BlockingMyClass.my_static_method() == 43
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY

    t0 = time.time()
    assert BlockingMyClass.my_class_method() == 44
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY

    assert list(z for z in obj) == list(range(42))


def test_class_sync_futures(synchronizer):
    BlockingMyClass = synchronizer.create_blocking(MyClass)
    BlockingBase = synchronizer.create_blocking(Base)
    assert BlockingMyClass.__name__ == "BlockingMyClass"
    obj = BlockingMyClass(x=42)
    assert isinstance(obj, BlockingMyClass)
    assert isinstance(obj, BlockingBase)
    obj.start()
    fut = obj.get_result(_future=True)
    assert isinstance(fut, concurrent.futures.Future)
    assert fut.result() == 1764

    t0 = time.monotonic()
    with obj as z:
        assert z == 42
        assert SLEEP_DELAY <= time.monotonic() - t0 < 2 * SLEEP_DELAY

    assert time.monotonic() - t0 >= 2 * SLEEP_DELAY


@pytest.mark.asyncio
async def test_class_async_as_method_attribute(synchronizer):
    BlockingMyClass = synchronizer.create_blocking(MyClass)
    BlockingBase = synchronizer.create_blocking(Base)
    assert BlockingMyClass.__name__ == "BlockingMyClass"
    obj = BlockingMyClass(x=42)
    assert isinstance(obj, BlockingMyClass)
    assert isinstance(obj, BlockingBase)
    await obj.start.aio()
    coro = obj.get_result.aio()
    assert inspect.iscoroutine(coro)
    assert await coro == 1764

    t0 = time.time()
    async with obj as z:
        assert z == 42
        assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY

    assert time.time() - t0 > 2 * SLEEP_DELAY

    lst = []

    async for z in obj:
        lst.append(z)

    assert lst == list(range(42))

    assert await obj.my_static_method.aio() == 43
    assert await obj.my_class_method.aio() == 44


@pytest.mark.skip(reason="Skip this until we've made it impossible to re-synchronize objects")
def test_event_loop(synchronizer):
    t0 = time.time()
    f_s = synchronizer.create_blocking(f)
    assert f_s(42) == 42 * 42
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY
    assert synchronizer._thread.is_alive()
    assert synchronizer._loop.is_running()
    synchronizer._close_loop()
    assert not synchronizer._loop.is_running()
    assert not synchronizer._thread.is_alive()

    new_loop = asyncio.new_event_loop()
    synchronizer._start_loop(new_loop)
    assert synchronizer._loop == new_loop
    assert synchronizer._loop.is_running()
    assert synchronizer._thread.is_alive()

    # Starting a loop again before closing throws.
    with pytest.raises(Exception):
        synchronizer._start_loop(new_loop)


def test_doc_transfer(synchronizer):
    class Foo:
        """Hello"""

        async def foo(self):
            """hello"""

    output_class = synchronizer.create_blocking(Foo)

    assert output_class.__doc__ == "Hello"
    assert output_class.foo.__doc__ == "hello"
    assert output_class.foo.aio.__doc__ == "hello"


def test_set_function_name(synchronizer):
    f_s = synchronizer.create_blocking(f, "xyz")
    assert f_s(42) == 1764
    assert f_s.__name__ == "xyz"


def test_set_class_name(synchronizer):
    BlockingBase = synchronizer.create_blocking(Base, "XYZBase")
    assert BlockingBase.__name__ == "XYZBase"
    BlockingMyClass = synchronizer.create_blocking(MyClass, "XYZMyClass")
    assert BlockingMyClass.__name__ == "XYZMyClass"


@pytest.mark.asyncio
async def test_blocking_nested_aio_returns_blocking_obj(synchronizer):
    class Foo:
        async def get_self(self):
            return self

    BlockingFoo = synchronizer.create_blocking(Foo)

    original = BlockingFoo()
    assert original.get_self() == original

    self_from_aio_interface = await original.get_self.aio()
    assert self_from_aio_interface == original
    assert isinstance(self_from_aio_interface, BlockingFoo)


def test_no_input_translation(monkeypatch, synchronizer):
    @synchronizer.create_blocking
    def does_input_translation(arg: float) -> str:
        return str(arg)

    @synchronizer.create_blocking
    @synchronizer.no_input_translation
    async def without_input_translation(arg: float) -> str:
        return str(arg)

    in_translate_spy = MagicMock(wraps=synchronizer._translate_scalar_in)
    monkeypatch.setattr(synchronizer, "_translate_scalar_in", in_translate_spy)
    does_input_translation(3.14)  # test without decorator, this *should* do input translation
    in_translate_spy.assert_called_once_with(3.14)

    in_translate_spy.reset_mock()
    without_input_translation(3.14)
    in_translate_spy.assert_not_called()


def test_no_output_translation(monkeypatch, synchronizer):
    @synchronizer.create_blocking
    def does_output_translation(arg: float) -> str:
        return str(arg)

    @synchronizer.create_blocking
    @synchronizer.no_output_translation
    async def without_output_translation(arg: float) -> str:
        return str(arg)

    out_translate_spy = MagicMock(wraps=synchronizer._translate_scalar_out)
    monkeypatch.setattr(synchronizer, "_translate_scalar_out", out_translate_spy)
    does_output_translation(3.14)  # test without decorator, this *should* do input translation
    out_translate_spy.assert_called_once_with("3.14")

    out_translate_spy.reset_mock()
    without_output_translation(3.14)
    out_translate_spy.assert_not_called()


@pytest.mark.asyncio
async def test_non_async_aiter(synchronizer):
    async def some_async_gen():
        yield "foo"
        yield "bar"

    class It:
        def __aiter__(self):
            self._gen = some_async_gen()
            return self

        async def __anext__(self):
            value = await self._gen.__anext__()
            return value

        async def aclose(self):
            await self._gen.aclose()

    WrappedIt = synchronizer.create_blocking(It, name="WrappedIt")

    # just a sanity check of the original iterable:
    orig_async_it = It()
    assert [v async for v in orig_async_it] == ["foo", "bar"]
    await orig_async_it.aclose()

    # check async iteration on the wrapped iterator
    it = WrappedIt()
    assert [v async for v in it] == ["foo", "bar"]
    await it.aclose()

    # check sync iteration on the wrapped iterator
    it = WrappedIt()
    assert list(it) == ["foo", "bar"]
    it.close()


def test_generic_baseclass():
    T = typing.TypeVar("T")
    V = typing.TypeVar("V")

    class GenericClass(typing.Generic[T, V]):
        async def do_something(self):
            return 1

    s = synchronicity.Synchronizer()
    WrappedGenericClass = s.create_blocking(GenericClass, name="BlockingGenericClass")

    assert WrappedGenericClass[str, float].__args__ == (str, float)

    instance: WrappedGenericClass[str, float] = WrappedGenericClass()  #  should be allowed
    assert isinstance(instance, WrappedGenericClass)
    assert instance.do_something() == 1

    Q = typing.TypeVar("Q")
    Y = typing.TypeVar("Y")

    class GenericSubclass(GenericClass[Q, Y]):
        pass

    WrappedGenericSubclass = s.create_blocking(GenericSubclass, name="BlockingGenericSubclass")
    assert WrappedGenericSubclass[bool, int].__args__ == (bool, int)
    instance_2 = WrappedGenericSubclass()
    assert isinstance(instance_2, WrappedGenericSubclass)
    assert isinstance(instance_2, WrappedGenericClass)  # still instance of parent
    assert instance.do_something() == 1  # has base methods


@pytest.mark.asyncio
async def test_async_cancellation(synchronizer):
    states = []

    async def foo(abort_cancellation: bool, cancel_self: bool = False):
        states.append("ready")
        if cancel_self:
            asyncio.tasks.current_task().cancel()
        try:
            await asyncio.sleep(10)
        except asyncio.CancelledError:
            states.append("cancelled")
            await asyncio.sleep(0.1)
            states.append("handled cancellation")
            if not abort_cancellation:
                raise
        return "done"

    wrapped_foo = synchronizer.create_blocking(foo)

    async def start_task(abort_cancellation: bool, cancel_self: bool = False):
        states.clear()
        calling_task = asyncio.create_task(
            wrapped_foo.aio(abort_cancellation=abort_cancellation, cancel_self=cancel_self)
        )
        while "ready" not in states:
            await asyncio.sleep(0.01)  # do't cancel before the task even starts
        return calling_task

    # Case 1: cancel in parent goes into the coroutine and comes back out:
    calling_task = await start_task(abort_cancellation=False)
    calling_task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await calling_task
    assert states == ["ready", "cancelled", "handled cancellation"]

    # Case 2: cancel in parent goes into the coroutine and is "aborted" by the coroutine:
    calling_task = await start_task(abort_cancellation=True)
    calling_task.cancel()
    assert await calling_task == "done"
    assert states == ["ready", "cancelled", "handled cancellation"]

    # Case 3: cancellation from within the coroutine itself comes back out:
    calling_task = await start_task(abort_cancellation=False, cancel_self=True)
    with pytest.raises(asyncio.CancelledError):
        await calling_task
    assert states == ["ready", "cancelled", "handled cancellation"]

    # Case 4: cancellation of the synchronicity task containing the coroutine itself
    # but it's caught and should not be propagated to the caller:
    calling_task = await start_task(abort_cancellation=True, cancel_self=True)
    assert await calling_task == "done"
    assert "ready" in states
    assert states == ["ready", "cancelled", "handled cancellation"]
